/*	$Id: cpu_mem.c,v 1.2 2003/07/22 15:09:43 nonaka Exp $	*/

/*
 * Copyright (c) 2002-2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "cpu.h"


BYTE *cpumem = 0;
DWORD extmem_size = 0;
BYTE protectmem_size = 0;

int
init_cpumem(BYTE usemem)
{
	DWORD size;

	if (usemem > 13)
		usemem = 13;
	size = usemem << 20;

	if (extmem_size != size - (LOWMEM - 0x100000)) {
		if (cpumem) {
			free(cpumem);
			cpumem = 0;
		}

		if (size <= LOWMEM - 0x100000) {
			extmem_size = 0;
			cpumem = 0;
		} else {
			extmem_size = size - (LOWMEM - 0x100000);
			cpumem = (LPBYTE)malloc(extmem_size);
			if (cpumem == NULL) {
				protectmem_size = 0;
				return FAILURE;
			}
		}
	}

	if (cpumem)
		memset(cpumem, 0, extmem_size);

	protectmem_size = usemem;
	return SUCCESS;
}


/*
 * virtual address -> physical address
 */
BYTE MEM_CALL
cpu_vmemoryread(int idx, DWORD madr)
{
	DWORD paddr;

	ASSERT(idx >= 0);

	if (!CPU_INST_AS32)
		madr &= 0xffff;
#if 1
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* real or vm86 mode */
		paddr = CPU_STAT_SREGBASE(idx) + madr;
	} else {
		/* protected mode */
		paddr = 0;
		assert(0);
	}
#else
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* real or vm86 mode */
	} else {
		/* protected mode */
	}

	/* limit check */
	if (idx != CPU_SS_INDEX) {
		if (madr > CPU_STAT_SREG(idx).limit) {
			ia32_panic(": segment limit check failed. "
				   "(sreg = %d, madr = %08x, limit = %08x)",
				   idx, madr, CPU_STAT_SREG(idx).limit);
		}
	} else {
		if (madr < CPU_STAT_SREG(idx).limit) {
			ia32_panic(": segment limit check failed. "
				   "(sreg = %d, madr = %08x, limit = %08x)",
				   idx, madr, CPU_STAT_SREG(idx).limit);
		}
	}
	paddr = CPU_STAT_SREGBASE(idx) + madr;

	if (CPU_INST_PAGING) {
		/* convert linear addr to phys addr */
	}
#endif
	return cpu_memoryread(paddr);
}

WORD MEM_CALL
cpu_vmemoryread_w(int idx, DWORD madr)
{
	DWORD paddr;

	ASSERT(idx >= 0);

	if (!CPU_INST_AS32)
		madr &= 0xffff;
#if 1
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* real or vm86 mode */
		paddr = CPU_STAT_SREGBASE(idx) + madr;
	} else {
		/* protected mode */
		paddr = 0;
		assert(0);
	}
#else
#endif
	return cpu_memoryread_w(paddr);
}

DWORD MEM_CALL
cpu_vmemoryread_d(int idx, DWORD madr)
{
	DWORD paddr;

	ASSERT(idx >= 0);

	if (!CPU_INST_AS32)
		madr &= 0xffff;
#if 1
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* real or vm86 mode */
		paddr = CPU_STAT_SREGBASE(idx) + madr;
	} else {
		/* protected mode */
		paddr = 0;
		assert(0);
	}
#else
#endif
	return cpu_memoryread_d(paddr);
}

/* vaddr memory write */
void MEM_CALL
cpu_vmemorywrite(int idx, DWORD madr, BYTE val)
{
	DWORD paddr;

	ASSERT(idx >= 0);

	if (!CPU_INST_AS32)
		madr &= 0xffff;
#if 1
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* real or vm86 mode */
		paddr = CPU_STAT_SREGBASE(idx) + madr;
	} else {
		/* protected mode */
		paddr = 0;
		assert(0);
	}
#else
#endif
	return cpu_memorywrite(paddr, val);
}

void MEM_CALL
cpu_vmemorywrite_w(int idx, DWORD madr, WORD val)
{
	DWORD paddr;

	ASSERT(idx >= 0);

	if (!CPU_INST_AS32)
		madr &= 0xffff;
#if 1
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* real or vm86 mode */
		paddr = CPU_STAT_SREGBASE(idx) + madr;
	} else {
		/* protected mode */
		paddr = 0;
		assert(0);
	}
#else
#endif
	return cpu_memorywrite_w(paddr, val);
}

void MEM_CALL
cpu_vmemorywrite_d(int idx, DWORD madr, DWORD val)
{
	DWORD paddr;

	ASSERT(idx >= 0);

	if (!CPU_INST_AS32)
		madr &= 0xffff;
#if 1
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* real or vm86 mode */
		paddr = CPU_STAT_SREGBASE(idx) + madr;
	} else {
		/* protected mode */
		paddr = 0;
		assert(0);
	}
#else
#endif
	return cpu_memorywrite_d(paddr, val);
}


/*
 * physical address memory function
 */
void MEM_CALL
cpu_memorywrite_d(DWORD address, DWORD value)
{
	DWORD adr = address & cpumem_addrmask;

	if (adr < LOWMEM - 3) {
		__i286_memorywrite_d(adr, value);
	} else if (adr < LOWMEM) {
		cpu_memorywrite_w(adr, value & 0xffff);
		cpu_memorywrite_w(adr + 2, (value >> 16) & 0xffff);
	} else {
		adr -= LOWMEM;
		if (adr < extmem_size) {
			SETDWORD(cpumem + adr, value);
		}
	}
}

void MEM_CALL
cpu_memorywrite_w(DWORD address, WORD value)
{
	DWORD adr = address & cpumem_addrmask;

	if (adr < LOWMEM - 1) {
		__i286_memorywrite_w(adr, value);
	} else if (adr < LOWMEM) {
		__i286_memorywrite(adr, value & 0xff);
		cpumem[adr - (LOWMEM - 1)] = (value >> 8) & 0xff;
	} else {
		adr -= LOWMEM;
		if (adr < extmem_size) {
			SETWORD(cpumem + adr, value);
		}
	}
}

void MEM_CALL
cpu_memorywrite(DWORD address, BYTE value)
{
	DWORD adr = address & cpumem_addrmask;

	if (adr < LOWMEM) {
		__i286_memorywrite(adr, value);
	} else {
		adr -= LOWMEM;
		if (adr < extmem_size) {
			cpumem[adr] = value;
		}
	}
}

DWORD MEM_CALL
cpu_memoryread_d(DWORD address)
{
	DWORD adr = address & cpumem_addrmask;
	DWORD val;

	if (adr < LOWMEM - 3) {
		val = __i286_memoryread_d(adr);
	} else if (adr < LOWMEM) {
		val = cpu_memoryread_w(adr);
		val |= (DWORD)cpu_memoryread_w(adr + 2) << 16;
	} else {
		adr -= LOWMEM;
		if (adr < extmem_size) {
			val = GETDWORD(cpumem + adr);
		} else {
			val = (DWORD)-1;
		}
	}

	return val;
}

WORD MEM_CALL
cpu_memoryread_w(DWORD address)
{
	DWORD adr = address & cpumem_addrmask;
	WORD val;

	if (adr < LOWMEM - 1) {
		val = __i286_memoryread_w(adr);
	} else if (adr < LOWMEM) {
		val = cpu_memoryread(adr);
		val |= (WORD)cpumem[adr - (LOWMEM - 1)] << 8;
	} else {
		adr -= LOWMEM;
		if (adr < extmem_size) {
			val = GETWORD(cpumem + adr);
		} else {
			val = (WORD)-1;
		}
	}

	return val;
}

BYTE MEM_CALL
cpu_memoryread(DWORD address)
{
	DWORD adr = address & cpumem_addrmask;
	BYTE val;

	if (adr < LOWMEM) {
		val = __i286_memoryread(adr);
	} else {
		adr -= LOWMEM;
		if (adr < extmem_size) {
			val = cpumem[adr];
		} else {
			val = (BYTE)-1;
		}
	}

	return val;
}
