/*	$Id: exception.c,v 1.1 2003/07/22 15:09:43 nonaka Exp $	*/

/*
 * Copyright (c) 2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "cpu.h"
#include "ia32.mcr"

void
memory_dump(int idx, DWORD madr)
{
	DWORD addr;
	size_t size;
	unsigned char buf[16];
	size_t s, i;
	BYTE p;

	if (madr < 0x80) {
		size = madr + 0x80;
		addr = 0;
	} else {
		size = 0x100;
		addr = madr - 0x80;
	}
	printf("memory dump\n-- \n");
	for (s = 0; s < size; s++) {
		if ((s % 16) == 0) {
			printf("%08x: ", addr + s);
			memset(buf, '.', sizeof(buf));
		}

		p = cpu_vmemoryread(idx, addr + s);
		printf("%02x ", p);
		if (p >= 0x20 && p <= 0x7e)
			buf[s % 16] = p;

		if ((s % 16) == 15) {
			printf("| ");
			for (i = 0; i < sizeof(buf); i++)
				printf("%c", buf[i]);
			printf("\n");
		}
	}
}

void
exception(int num, int vec)
{

	printf("exception: num = %d, vec = %d\n", num, vec);
	put_cpuinfo();
	memory_dump(CPU_CS_INDEX, CPU_PREV_EIP);
	memory_dump(CPU_SS_INDEX, CPU_ESP);
	assert(0);
}

void
interrupt(int num, int softintp, int errorp, int error_code)
{
	BYTE op;

	UNUSED(softintp);
	UNUSED(errorp);
	UNUSED(error_code);

	op = cpu_vmemoryread(CPU_CS_INDEX, CPU_EIP);
	if (op == 0xf4) {	/* hlt */
		CPU_EIP++;
	}

	/* XXX REAL MODE */

#if defined(USE_ASM)
	REGPUSH0(CPU_FLAG);
#else
	REGPUSH0(REAL_FLAGREG);
#endif
	REGPUSH0(CPU_CS);
	REGPUSH0(CPU_IP);

	CPU_FLAG &= ~(T_FLAG | I_FLAG);
	CPU_TRAP = 0;

	CPU_EIP = GETWORD(mem + num * 4);
#if defined(USE_ASM)
	CPU_CS = GETWORD(mem + num * 4 + 2);
	CS_BASE = (DWORD)(CPU_CS << 4);
#else
	CPU_SET_SEGREG(CPU_CS_INDEX, GETWORD(mem + num * 4 + 2));
#endif
	CPU_CLOCK(20);
}
