/*	$Id: bit_byte.c,v 1.2 2003/07/22 15:09:57 nonaka Exp $	*/

/*
 * Copyright (c) 2002-2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "cpu.h"
#include "ia32.mcr"

#include "bit_byte.h"


/*
 * BT
 */
void
BT_EwGw(void)
{
	WORD *out;
	DWORD op, src, dst, madr, ad;

	PREPART_EA_REG16(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg16_b20[op];
		dst = *out;
		if (dst & (1 << (src & 0x0f))) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	} else {
		CPU_CLOCK(7);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		dst = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, ad + (2 * (src >> 4)));
		if (dst & (1 << (src & 0x0f))) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	}
}

void
BT_EdGd(void)
{
	DWORD *out;
	DWORD op, src, dst, madr, ad;

	PREPART_EA_REG32(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg32_b20[op];
		dst = *out;
		if (dst & (1 << (src & 0x1f))) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	} else {
		CPU_CLOCK(7);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr);
		dst = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, ad + (4 * (src >> 5)));
		if (dst & (1 << (src & 0x1f))) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	}
}

void
BT_EwIb(DWORD op)
{
	WORD *out;
	DWORD src, dst, madr, ad;

	GET_PCBYTE(src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg16_b20[op];
		dst = *out;
		if (dst & (1 << (src & 0x0f))) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	} else {
		CPU_CLOCK(6);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		dst = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, ad + (2 * (src >> 4)));
		if (dst & (1 << (src & 0x0f))) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	}
}

void
BT_EdIb(DWORD op)
{
	DWORD *out;
	DWORD src, dst, madr, ad;

	GET_PCBYTE(src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg32_b20[op];
		dst = *out;
		if (dst & (1 << (src & 0x1f))) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	} else {
		CPU_CLOCK(6);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr);
		dst = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, ad + (4 * (src >> 5)));
		if (dst & (1 << (src & 0x1f))) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	}
}

/*
 * BTS
 */
void
BTS_EwGw(void)
{
	WORD *out;
	DWORD op, src, dst, madr, ad;
	DWORD bit;

	PREPART_EA_REG16(op, src);
	bit = 1 << (src & 0x0f);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg16_b20[op];
		dst = *out;
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
			dst |= bit;
			*out = dst;
		}
	} else {
		CPU_CLOCK(7);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		ad += 2 * (src >> 4);
		dst = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, ad);
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
			dst |= bit;
			cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, ad, dst);
		}
	}
}

void
BTS_EdGd(void)
{
	DWORD *out;
	DWORD op, src, dst, madr, ad;
	DWORD bit;

	PREPART_EA_REG32(op, src);
	bit = 1 << (src & 0x1f);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg32_b20[op];
		dst = *out;
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
			dst |= bit;
			*out = dst;
		}
	} else {
		CPU_CLOCK(7);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr);
		ad += 4 * (src >> 5);
		dst = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, ad);
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
			dst |= bit;
			cpu_vmemorywrite_d(CPU_INST_SEGREG_INDEX, ad, dst);
		}
	}
}

void
BTS_EwIb(DWORD op)
{
	WORD *out;
	DWORD src, dst, madr, ad;
	DWORD bit;

	GET_PCBYTE(src);
	bit = 1 << (src & 0x0f);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg16_b20[op];
		dst = *out;
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
			dst |= bit;
			*out = dst;
		}
	} else {
		CPU_CLOCK(6);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		ad += 2 * (src >> 4);
		dst = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, ad);
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
			dst |= bit;
			cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, ad, dst);
		}
	}
}

void
BTS_EdIb(DWORD op)
{
	DWORD *out;
	DWORD src, dst, madr, ad;
	DWORD bit;

	GET_PCBYTE(src);
	bit = 1 << (src & 0x1f);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg32_b20[op];
		dst = *out;
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
			dst |= bit;
			*out = dst;
		}
	} else {
		CPU_CLOCK(6);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr);
		ad += 4 * (src >> 5);
		dst = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, ad);
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
			dst |= bit;
			cpu_vmemorywrite_d(CPU_INST_SEGREG_INDEX, ad, dst);
		}
	}
}

/*
 * BTR
 */
void
BTR_EwGw(void)
{
	WORD *out;
	DWORD op, src, dst, madr, ad;
	DWORD bit;

	PREPART_EA_REG16(op, src);
	bit = 1 << (src & 0x0f);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg16_b20[op];
		dst = *out;
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
			dst &= ~bit;
			*out = dst;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	} else {
		CPU_CLOCK(7);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		ad += 2 * (src >> 4);
		dst = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, ad);
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
			dst &= ~bit;
			cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, ad, dst);
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	}
}

void
BTR_EdGd(void)
{
	DWORD *out;
	DWORD op, src, dst, madr, ad;
	DWORD bit;

	PREPART_EA_REG32(op, src);
	bit = 1 << (src & 0x1f);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg32_b20[op];
		dst = *out;
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
			dst &= ~bit;
			*out = dst;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	} else {
		CPU_CLOCK(7);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr);
		ad += 4 * (src >> 5);
		dst = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, ad);
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
			dst &= ~bit;
			cpu_vmemorywrite_d(CPU_INST_SEGREG_INDEX, ad, dst);
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	}
}

void
BTR_EwIb(DWORD op)
{
	WORD *out;
	DWORD src, dst, madr, ad;
	DWORD bit;

	GET_PCBYTE(src);
	bit = 1 << (src & 0x0f);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg16_b20[op];
		dst = *out;
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
			dst &= ~bit;
			*out = dst;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	} else {
		CPU_CLOCK(6);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		ad += 2 * (src >> 4);
		dst = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, ad);
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
			dst &= ~bit;
			cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, ad, dst);
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	}
}

void
BTR_EdIb(DWORD op)
{
	DWORD *out;
	DWORD src, dst, madr, ad;
	DWORD bit;

	GET_PCBYTE(src);
	bit = 1 << (src & 0x1f);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg32_b20[op];
		dst = *out;
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
			dst &= ~bit;
			*out = dst;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	} else {
		CPU_CLOCK(6);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr);
		ad += 4 * (src >> 5);
		dst = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, ad);
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
			dst &= ~bit;
			cpu_vmemorywrite_d(CPU_INST_SEGREG_INDEX, ad, dst);
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
	}
}


/*
 * BTC
 */
void
BTC_EwGw(void)
{
	WORD *out;
	DWORD op, src, dst, madr, ad;
	DWORD bit;

	PREPART_EA_REG16(op, src);
	bit = 1 << (src & 0x0f);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg16_b20[op];
		dst = *out;
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
		dst ^= bit;
		*out = dst;
	} else {
		CPU_CLOCK(7);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		ad += 2 * (src >> 4);
		dst = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, ad);
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
		dst ^= bit;
		cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, ad, dst);
	}
}

void
BTC_EdGd(void)
{
	DWORD *out;
	DWORD op, src, dst, madr, ad;
	DWORD bit;

	PREPART_EA_REG32(op, src);
	bit = 1 << (src & 0x1f);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg32_b20[op];
		dst = *out;
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
		dst ^= bit;
		*out = dst;
	} else {
		CPU_CLOCK(7);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr);
		ad += 4 * (src >> 5);
		dst = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, ad);
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
		dst ^= bit;
		cpu_vmemorywrite_d(CPU_INST_SEGREG_INDEX, ad, dst);
	}
}

void
BTC_EwIb(DWORD op)
{
	WORD *out;
	DWORD src, dst, madr, ad;
	DWORD bit;

	GET_PCBYTE(src);
	bit = 1 << (src & 0x0f);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg16_b20[op];
		dst = *out;
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
		dst ^= bit;
		*out = dst;
	} else {
		CPU_CLOCK(6);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		ad += 2 * (src >> 4);
		dst = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, ad);
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
		dst ^= bit;
		cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, ad, dst);
	}
}

void
BTC_EdIb(DWORD op)
{
	DWORD *out;
	DWORD src, dst, madr, ad;
	DWORD bit;

	GET_PCBYTE(src);
	bit = 1 << (src & 0x1f);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg32_b20[op];
		dst = *out;
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
		dst ^= bit;
		*out = dst;
	} else {
		CPU_CLOCK(6);
		madr = calc_ea_dst(op);
		ad = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr);
		ad += 4 * (src >> 5);
		dst = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, ad);
		if (dst & bit) {
			CPU_FLAGL |= C_FLAG;
		} else {
			CPU_FLAGL &= ~C_FLAG;
		}
		dst ^= bit;
		cpu_vmemorywrite_d(CPU_INST_SEGREG_INDEX, ad, dst);
	}
}

/*
 * BSF
 */
void
BSF_GwEw(void)
{
	WORD *out;
	DWORD op, src, bit;

	PREPART_REG16_EA(op, src, out, 2, 7);
	if (src == 0) {
		CPU_FLAGL |= Z_FLAG;
		/* dest reg is undefined */
	} else {
		CPU_FLAGL &= ~Z_FLAG;
		for (bit = 0; !(src & (1 << bit)); bit++)
			continue;
		*out = bit;
	}
}

void
BSF_GdEd(void)
{
	DWORD *out;
	DWORD op, src, bit;

	PREPART_REG32_EA(op, src, out, 2, 7);
	if (src == 0) {
		CPU_FLAGL |= Z_FLAG;
		/* dest reg is undefined */
	} else {
		CPU_FLAGL &= ~Z_FLAG;
		for (bit = 0; !(src & (1 << bit)); bit++)
			continue;
		*out = bit;
	}
}

/*
 * BSR
 */
void
BSR_GwEw(void)
{
	WORD *out;
	DWORD op, src, bit;

	PREPART_REG16_EA(op, src, out, 2, 7);
	if (src == 0) {
		CPU_FLAGL |= Z_FLAG;
		/* dest reg is undefined */
	} else {
		CPU_FLAGL &= ~Z_FLAG;
		for (bit = 15; !(src & (1 << bit)); bit--)
			continue;
		*out = bit;
	}
}

void
BSR_GdEd(void)
{
	DWORD *out;
	DWORD op, src, bit;

	PREPART_REG32_EA(op, src, out, 2, 7);
	if (src == 0) {
		CPU_FLAGL |= Z_FLAG;
		/* dest reg is undefined */
	} else {
		CPU_FLAGL &= ~Z_FLAG;
		for (bit = 31; !(src & (1 << bit)); bit--)
			continue;
		*out = bit;
	}
}

/*
 * SETcc
 */
void
SETO_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if (!CPU_OV)  {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if (!CPU_OV) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETNO_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if (CPU_OV) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if (CPU_OV) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETC_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if (!(CPU_FLAGL & C_FLAG)) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if (!(CPU_FLAGL & C_FLAG)) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETNC_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if (CPU_FLAGL & C_FLAG) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if (CPU_FLAGL & C_FLAG) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETZ_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if (!(CPU_FLAGL & Z_FLAG)) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if (!(CPU_FLAGL & Z_FLAG)) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETNZ_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if (CPU_FLAGL & Z_FLAG) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if (CPU_FLAGL & Z_FLAG) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETA_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if (CPU_FLAGL & (Z_FLAG | C_FLAG)) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if (CPU_FLAGL & (Z_FLAG | C_FLAG)) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETNA_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if (!(CPU_FLAGL & (Z_FLAG | C_FLAG))) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if (!(CPU_FLAGL & (Z_FLAG | C_FLAG))) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETS_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if (!(CPU_FLAGL & S_FLAG)) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if (!(CPU_FLAGL & S_FLAG)) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETNS_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if (CPU_FLAGL & S_FLAG) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if (CPU_FLAGL & S_FLAG) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETP_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if (!(CPU_FLAGL & P_FLAG)) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if (!(CPU_FLAGL & P_FLAG)) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETNP_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if (CPU_FLAGL & P_FLAG) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if (CPU_FLAGL & P_FLAG) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETL_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if ((!CPU_OV) == (!(CPU_FLAGL & S_FLAG))) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if ((!CPU_OV) == (!(CPU_FLAGL & S_FLAG))) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETNL_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if ((!CPU_OV) != (!(CPU_FLAGL & S_FLAG))) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if ((!CPU_OV) != (!(CPU_FLAGL & S_FLAG))) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETLE_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if ((!(CPU_FLAGL & Z_FLAG)) && ((!(CPU_FLAGL & S_FLAG)) == !CPU_OV)) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if ((!(CPU_FLAGL & Z_FLAG)) && ((!(CPU_FLAGL & S_FLAG)) == !CPU_OV)) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

void
SETNLE_Eb(void)
{
	BYTE *out;
	DWORD op, src, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		if ((CPU_FLAGL & Z_FLAG) || ((!(CPU_FLAGL & S_FLAG)) != !CPU_OV)) {
			*out = 0;
		} else {
			*out = 1;
		}
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		if ((CPU_FLAGL & Z_FLAG) || ((!(CPU_FLAGL & S_FLAG)) != !CPU_OV)) {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 0);
		} else {
			cpu_vmemorywrite(CPU_INST_SEGREG_INDEX, madr, 1);
		}
	}
}

/*
 * TEST
 */
void
TEST_EbGb(void)
{
	BYTE *out;
	DWORD op, src, tmp, madr;

	PREPART_EA_REG8(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		tmp = *out;
		ANDBYTE(tmp, src);
	} else {
		CPU_CLOCK(7);
		madr = calc_ea_dst(op);
		tmp = cpu_vmemoryread(CPU_INST_SEGREG_INDEX, madr);
		ANDBYTE(tmp, src);
	}
}

void
TEST_EwGw(void)
{
	WORD *out;
	DWORD op, src, tmp, madr;

	PREPART_EA_REG16(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg16_b20[op];
		tmp = *out;
		ANDWORD(tmp, src);
	} else {
		CPU_CLOCK(7);
		madr = calc_ea_dst(op);
		tmp = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		ANDWORD(tmp, src);
	}
}

void
TEST_EdGd(void)
{
	DWORD *out;
	DWORD op, src, tmp, madr;

	PREPART_EA_REG32(op, src);
	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg32_b20[op];
		tmp = *out;
		ANDWORD(tmp, src);
	} else {
		CPU_CLOCK(7);
		madr = calc_ea_dst(op);
		tmp = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr);
		ANDWORD(tmp, src);
	}
}

void
TEST_ALIb(void)
{
	DWORD src, tmp;

	CPU_CLOCK(3);
	GET_PCBYTE(src);
	tmp = CPU_AL;
	ANDBYTE(tmp, src);
}

void
TEST_AXIw(void)
{
	DWORD src, tmp;

	CPU_CLOCK(3);
	GET_PCWORD(src);
	tmp = CPU_AX;
	ANDWORD(tmp, src);
}

void
TEST_EAXId(void)
{
	DWORD src, tmp;

	CPU_CLOCK(3);
	GET_PCDWORD(src);
	tmp = CPU_EAX;
	ANDDWORD(tmp, src);
}

void
TEST_EbIb(DWORD op)
{
	BYTE *out;
	DWORD src, tmp, madr;

	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg8_b20[op];
		tmp = *out;
		GET_PCBYTE(src);
		ANDBYTE(tmp, src);
	} else {
		CPU_CLOCK(6);
		madr = calc_ea_dst(op);
		tmp = cpu_vmemoryread(CPU_INST_SEGREG_INDEX, madr);
		GET_PCBYTE(src);
		ANDBYTE(tmp, src);
	}
}

void
TEST_EwIw(DWORD op)
{
	WORD *out;
	DWORD src, tmp, madr;

	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg16_b20[op];
		tmp = *out;
		GET_PCWORD(src);
		ANDWORD(tmp, src);
	} else {
		CPU_CLOCK(6);
		madr = calc_ea_dst(op);
		tmp = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		GET_PCWORD(src);
		ANDWORD(tmp, src);
	}
}

void
TEST_EdId(DWORD op)
{
	DWORD *out;
	DWORD src, tmp, madr;

	if (op >= 0xc0) {
		CPU_CLOCK(2);
		out = reg32_b20[op];
		tmp = *out;
		GET_PCDWORD(src);
		ANDDWORD(tmp, src);
	} else {
		CPU_CLOCK(6);
		madr = calc_ea_dst(op);
		tmp = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr);
		GET_PCDWORD(src);
		ANDDWORD(tmp, src);
	}
}

