/*	$Id: resolve.c,v 1.3 2003/07/22 15:09:44 nonaka Exp $	*/

/*
 * Copyright (c) 2002-2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "cpu.h"
#include "ia32.mcr"


static WORD (*calc_ea_dst_tbl[0x100])(void);
static WORD (*calc_lea_tbl[0xc0])(void);
static DWORD (*calc_ea32_dst_tbl[0x100])(void);
static DWORD (*calc_lea32_tbl[0xc0])(void);


DWORD
calc_ea_dst(DWORD op)
{

	if (!CPU_INST_AS32)
		return (*calc_ea_dst_tbl[op])();
	return (*calc_ea32_dst_tbl[op])();
}

DWORD
calc_lea(DWORD op)
{

	if (!CPU_INST_AS32)
		return (*calc_lea_tbl[op])();
	return (*calc_lea32_tbl[op])();
}


/*
 * common
 */
WORD
ea_nop(void)
{

	CPU_INST_SEGREG_INDEX = DS_FIX;
	return 0;
}

DWORD
ea32_nop(void)
{

	CPU_INST_SEGREG_INDEX = DS_FIX;
	return 0;
}


/*
 * 16bit
 */

/*
 * ea_dest
 */
static WORD
ea_bx_si(void)
{

	CPU_INST_SEGREG_INDEX = DS_FIX;
	return (CPU_BX + CPU_SI);
}

static WORD
ea_bx_si_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return (adrs + CPU_BX + CPU_SI);
}

static WORD
ea_bx_si_disp16(void)
{
	DWORD adrs;

	GET_PCWORD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return (adrs + CPU_BX + CPU_SI);
}

static WORD
ea_bx_di(void)
{

	CPU_INST_SEGREG_INDEX = DS_FIX;
	return (CPU_BX + CPU_DI);
}

static WORD
ea_bx_di_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return (adrs + CPU_BX + CPU_DI);
}

static WORD
ea_bx_di_disp16(void)
{
	DWORD adrs;

	GET_PCWORD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return (adrs + CPU_BX + CPU_DI);
}

static WORD
ea_bp_si(void)
{

	CPU_INST_SEGREG_INDEX = SS_FIX;
	return (CPU_BP + CPU_SI);
}

static WORD
ea_bp_si_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = SS_FIX;
	return (adrs + CPU_BP + CPU_SI);
}

static WORD
ea_bp_si_disp16(void)
{
	DWORD adrs;

	GET_PCWORD(adrs);
	CPU_INST_SEGREG_INDEX = SS_FIX;
	return (adrs + CPU_BP + CPU_SI);
}

static WORD
ea_bp_di(void)
{

	CPU_INST_SEGREG_INDEX = SS_FIX;
	return (CPU_BP + CPU_DI);
}

static WORD
ea_bp_di_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = SS_FIX;
	return (adrs + CPU_BP + CPU_DI);
}

static WORD
ea_bp_di_disp16(void)
{
	DWORD adrs;

	GET_PCWORD(adrs);
	CPU_INST_SEGREG_INDEX = SS_FIX;
	return (adrs + CPU_BP + CPU_DI);
}

static WORD
ea_si(void)
{

	CPU_INST_SEGREG_INDEX = DS_FIX;
	return CPU_SI;
}

static WORD
ea_si_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return (adrs + CPU_SI);
}

static WORD
ea_si_disp16(void)
{
	DWORD adrs;

	GET_PCWORD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return (adrs + CPU_SI);
}

static WORD
ea_di(void)
{

	CPU_INST_SEGREG_INDEX = DS_FIX;
	return CPU_DI;
}

static WORD
ea_di_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return (adrs + CPU_DI);
}

static WORD
ea_di_disp16(void)
{
	DWORD adrs;

	GET_PCWORD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return (adrs + CPU_DI);
}

static WORD
ea_disp16(void)
{
	DWORD adrs;

	GET_PCWORD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs;
}

static WORD
ea_bp_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = SS_FIX;
	return (adrs + CPU_BP);
}

static WORD
ea_bp_disp16(void)
{
	DWORD adrs;

	GET_PCWORD(adrs);
	CPU_INST_SEGREG_INDEX = SS_FIX;
	return (adrs + CPU_BP);
}

static WORD
ea_bx(void)
{

	CPU_INST_SEGREG_INDEX = DS_FIX;
	return CPU_BX;
}

static WORD
ea_bx_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return (adrs + CPU_BX);
}

static WORD
ea_bx_disp16(void)
{
	DWORD adrs;

	GET_PCWORD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return (adrs + CPU_BX);
}

static WORD (*c_ea_dst_tbl[])(void) = {
	ea_bx_si,		ea_bx_di,
	ea_bp_si,		ea_bp_di,
	ea_si,			ea_di,
	ea_disp16,		ea_bx,
	ea_bx_si_disp8,		ea_bx_di_disp8,
	ea_bp_si_disp8,		ea_bp_di_disp8,
	ea_si_disp8,		ea_di_disp8,
	ea_bp_disp8,		ea_bx_disp8,
	ea_bx_si_disp16,	ea_bx_di_disp16,
	ea_bp_si_disp16,	ea_bp_di_disp16,
	ea_si_disp16,		ea_di_disp16,
	ea_bp_disp16,		ea_bx_disp16,
};

/*
 * lea
 */
static WORD 
lea_bx_si(void)
{

	return (CPU_BX + CPU_SI);
}

static WORD 
lea_bx_si_disp8(void)
{
	WORD adrs;

	GET_PCBYTES(adrs);
	return (adrs + CPU_BX + CPU_SI);
}

static WORD 
lea_bx_si_disp16(void)
{
	WORD adrs;

	GET_PCWORD(adrs);
	return (adrs + CPU_BX + CPU_SI);
}

static WORD 
lea_bx_di(void)
{

	return (CPU_BX + CPU_DI);
}

static WORD 
lea_bx_di_disp8(void)
{
	WORD adrs;

	GET_PCBYTES(adrs);
	return (adrs + CPU_BX + CPU_DI);
}

static WORD 
lea_bx_di_disp16(void)
{
	WORD adrs;

	GET_PCWORD(adrs);
	return (adrs + CPU_BX + CPU_DI);
}

static WORD 
lea_bp_si(void)
{

	return (CPU_BP + CPU_SI);
}

static WORD 
lea_bp_si_disp8(void)
{

	WORD	adrs;

	GET_PCBYTES(adrs);
	return (adrs + CPU_BP + CPU_SI);
}

static WORD 
lea_bp_si_disp16(void)
{
	WORD adrs;

	GET_PCWORD(adrs);
	return (adrs + CPU_BP + CPU_SI);
}

static WORD 
lea_bp_di(void)
{

	return (CPU_BP + CPU_DI);
}

static WORD 
lea_bp_di_disp8(void)
{
	WORD adrs;

	GET_PCBYTES(adrs);
	return (adrs + CPU_BP + CPU_DI);
}

static WORD 
lea_bp_di_disp16(void)
{
	WORD adrs;

	GET_PCWORD(adrs);
	return (adrs + CPU_BP + CPU_DI);
}

static WORD 
lea_si(void)
{

	return CPU_SI;
}

static WORD 
lea_si_disp8(void)
{
	WORD adrs;

	GET_PCBYTES(adrs);
	return (adrs + CPU_SI);
}

static WORD 
lea_si_disp16(void)
{
	WORD adrs;

	GET_PCWORD(adrs);
	return (adrs + CPU_SI);
}

static WORD 
lea_di(void)
{

	return CPU_DI;
}

static WORD 
lea_di_disp8(void)
{
	WORD adrs;

	GET_PCBYTES(adrs);
	return (adrs + CPU_DI);
}

static WORD 
lea_di_disp16(void)
{
	WORD adrs;

	GET_PCWORD(adrs);
	return (adrs + CPU_DI);
}

static WORD 
lea_disp16(void)
{
	WORD adrs;

	GET_PCWORD(adrs);
	return adrs;
}

static WORD 
lea_bp_disp8(void)
{
	WORD adrs;

	GET_PCBYTES(adrs);
	return (adrs + CPU_BP);
}

static WORD 
lea_bp_disp16(void)
{
	WORD adrs;

	GET_PCWORD(adrs);
	return (adrs + CPU_BP);
}

static WORD 
lea_bx(void)
{

	return CPU_BX;
}

static WORD 
lea_bx_disp8(void)
{
	WORD adrs;

	GET_PCBYTES(adrs);
	return (adrs + CPU_BX);
}

static WORD 
lea_bx_disp16(void)
{
	WORD adrs;

	GET_PCWORD(adrs);
	return (adrs + CPU_BX);
}

static WORD (*c_lea_tbl[])(void) = {
	lea_bx_si,		lea_bx_di,
	lea_bp_si,		lea_bp_di,
	lea_si,			lea_di,
	lea_disp16,		lea_bx,
	lea_bx_si_disp8,	lea_bx_di_disp8,
	lea_bp_si_disp8,	lea_bp_di_disp8,
	lea_si_disp8,		lea_di_disp8,
	lea_bp_disp8,		lea_bx_disp8,
	lea_bx_si_disp16,	lea_bx_di_disp16,
	lea_bp_si_disp16,	lea_bp_di_disp16,
	lea_si_disp16,		lea_di_disp16,
	lea_bp_disp16,		lea_bx_disp16
};


/*
 * 32bit
 */

/* ea_dest sib */
static DWORD
sib_eax(DWORD base, DWORD scale)
{

	return (CPU_EAX << scale) + base;
}

static DWORD
sib_ecx(DWORD base, DWORD scale)
{

	return (CPU_ECX << scale) + base;
}

static DWORD
sib_edx(DWORD base, DWORD scale)
{

	return (CPU_EDX << scale) + base;
}

static DWORD
sib_ebx(DWORD base, DWORD scale)
{

	return (CPU_EBX << scale) + base;
}

static DWORD
sib_none(DWORD base, DWORD scale)
{

	UNUSED(base);
	UNUSED(scale);
	return 0;
}

static DWORD
sib_ebp(DWORD base, DWORD scale)
{

	return (CPU_EBP << scale) + base;
}

static DWORD
sib_esi(DWORD base, DWORD scale)
{

	return (CPU_ESI << scale) + base;
}

static DWORD
sib_edi(DWORD base, DWORD scale)
{

	return (CPU_EDI << scale) + base;
}

static DWORD (*sib_table[])(DWORD, DWORD) = {
	sib_eax,	sib_ecx,
	sib_edx,	sib_ebx,
	sib_none,	sib_ebp,
	sib_esi,	sib_edi,
};

/*
 * ea_dest 32
 */
static DWORD
ea32_eax(void)
{

	CPU_INST_SEGREG_INDEX = DS_FIX;
	return CPU_EAX;
}

static DWORD
ea32_ecx(void)
{

	CPU_INST_SEGREG_INDEX = DS_FIX;
	return CPU_ECX;
}

static DWORD
ea32_edx(void)
{

	CPU_INST_SEGREG_INDEX = DS_FIX;
	return CPU_EDX;
}

static DWORD
ea32_ebx(void)
{

	CPU_INST_SEGREG_INDEX = DS_FIX;
	return CPU_EBX;
}

static DWORD
ea32_sib(void)
{
	DWORD op, dst;
	DWORD base, idx, scale;

	GET_PCBYTE(op);
	base = op & 7;
	idx = (op >> 3) & 7;
	scale = (op >> 6) & 3;

	switch (base) {
	case 4:
		CPU_INST_SEGREG_INDEX = SS_FIX;
		return (*sib_table[idx])(CPU_ESP, scale);

	case 5:
		GET_PCDWORD(dst);
		break;

	default:
		dst = CPU_REGS_DWORD(base);
		break;
	}
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return (*sib_table[idx])(dst, scale);
}

static DWORD
ea32_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs;
}

static DWORD
ea32_esi(void)
{

	CPU_INST_SEGREG_INDEX = DS_FIX;
	return CPU_ESI;
}

static DWORD
ea32_edi(void)
{

	CPU_INST_SEGREG_INDEX = DS_FIX;
	return CPU_EDI;
}

static DWORD
ea32_eax_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs + CPU_EAX;
}

static DWORD
ea32_ecx_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs + CPU_ECX;
}

static DWORD
ea32_edx_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs + CPU_EDX;
}

static DWORD
ea32_ebx_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs + CPU_EBX;
}

static DWORD
ea32_sib_disp8(void)
{
	DWORD op;
	DWORD adrs;

	GET_PCBYTE(op);
	GET_PCBYTESD(adrs);
	switch (op & 7) {
	case 0: case 1: case 2: case 3: case 6:
		CPU_INST_SEGREG_INDEX = DS_FIX;
		break;

	case 4: case 5:
		CPU_INST_SEGREG_INDEX = SS_FIX;
		break;
	}
	return (*sib_table[(op >> 3) & 7])(CPU_REGS_DWORD(op & 7), op >> 6) + adrs;
}

static DWORD
ea32_ebp_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = SS_FIX;
	return adrs + CPU_EBP;
}

static DWORD
ea32_esi_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs + CPU_ESI;
}

static DWORD
ea32_edi_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs + CPU_EDI;
}

static DWORD
ea32_eax_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs + CPU_EAX;
}

static DWORD
ea32_ecx_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs + CPU_ECX;
}

static DWORD
ea32_edx_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs + CPU_EDX;
}

static DWORD
ea32_ebx_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs + CPU_EBX;
}

static DWORD
ea32_sib_disp32(void)
{
	DWORD op;
	DWORD adrs;

	GET_PCBYTE(op);
	GET_PCDWORD(adrs);
	switch (op & 7) {
	case 0: case 1: case 2: case 3: case 6:
		CPU_INST_SEGREG_INDEX = DS_FIX;
		break;

	case 4: case 5:
		CPU_INST_SEGREG_INDEX = SS_FIX;
		break;
	}
	return (*sib_table[(op >> 3) & 7])(CPU_REGS_DWORD(op & 7), op >> 6) + adrs;
}

static DWORD
ea32_ebp_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	CPU_INST_SEGREG_INDEX = SS_FIX;
	return adrs + CPU_EBP;
}

static DWORD
ea32_esi_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs + CPU_ESI;
}

static DWORD
ea32_edi_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	CPU_INST_SEGREG_INDEX = DS_FIX;
	return adrs + CPU_EDI;
}

static DWORD (*c_ea32_dst_tbl[])(void) = {
	ea32_eax,		ea32_ecx, 
	ea32_edx,		ea32_ebx,
	ea32_sib,		ea32_disp32,
	ea32_esi,		ea32_edi,
	ea32_eax_disp8,		ea32_ecx_disp8, 
	ea32_edx_disp8,		ea32_ebx_disp8,
	ea32_sib_disp8,		ea32_ebp_disp8,
	ea32_esi_disp8,		ea32_edi_disp8,
	ea32_eax_disp32,	ea32_ecx_disp32, 
	ea32_edx_disp32,	ea32_ebx_disp32,
	ea32_sib_disp32,	ea32_ebp_disp32,
	ea32_esi_disp32,	ea32_edi_disp32,
};

/*
 * lea 32
 */
static DWORD
lea32_eax(void)
{

	return CPU_EAX;
}

static DWORD
lea32_ecx(void)
{

	return CPU_ECX;
}

static DWORD
lea32_edx(void)
{

	return CPU_EDX;
}

static DWORD
lea32_ebx(void)
{

	return CPU_EBX;
}

static DWORD
lea32_sib(void)
{
	DWORD op, dst;
	DWORD base, idx, scale;

	GET_PCBYTE(op);
	base = op & 7;
	idx = (op >> 3) & 7;
	scale = (op >> 6) & 3;

	switch (base) {
	case 5:
		GET_PCDWORD(dst);
		break;

	default:
		dst = CPU_REGS_DWORD(base);
		break;
	}
	return (*sib_table[idx])(dst, scale);
}

static DWORD
lea32_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	return adrs;
}

static DWORD
lea32_esi(void)
{

	return CPU_ESI;
}

static DWORD
lea32_edi(void)
{

	return CPU_EDI;
}

static DWORD
lea32_eax_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	return adrs + CPU_EAX;
}

static DWORD
lea32_ecx_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	return adrs + CPU_ECX;
}

static DWORD
lea32_edx_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	return adrs + CPU_EDX;
}

static DWORD
lea32_ebx_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	return adrs + CPU_EBX;
}

static DWORD
lea32_sib_disp8(void)
{
	DWORD op;
	DWORD adrs;

	GET_PCBYTE(op);
	GET_PCBYTESD(adrs);
	return (*sib_table[(op >> 3) & 7])(CPU_REGS_DWORD(op & 7), op >> 6) + adrs;
}

static DWORD
lea32_ebp_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	return adrs + CPU_EBP;
}

static DWORD
lea32_esi_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	return adrs + CPU_ESI;
}

static DWORD
lea32_edi_disp8(void)
{
	DWORD adrs;

	GET_PCBYTESD(adrs);
	return adrs + CPU_EDI;
}

static DWORD
lea32_eax_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	return adrs + CPU_EAX;
}

static DWORD
lea32_ecx_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	return adrs + CPU_ECX;
}

static DWORD
lea32_edx_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	return adrs + CPU_EDX;
}

static DWORD
lea32_ebx_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	return adrs + CPU_EBX;
}

static DWORD
lea32_sib_disp32(void)
{
	DWORD op;
	DWORD adrs;

	GET_PCBYTE(op);
	GET_PCDWORD(adrs);
	return (*sib_table[(op >> 3) & 7])(CPU_REGS_DWORD(op & 7), op >> 6) + adrs;
}

static DWORD
lea32_ebp_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	return adrs + CPU_EBP;
}

static DWORD
lea32_esi_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	return adrs + CPU_ESI;
}

static DWORD
lea32_edi_disp32(void)
{
	DWORD adrs;

	GET_PCDWORD(adrs);
	return adrs + CPU_EDI;
}

static DWORD (*c_lea32_tbl[])(void) = {
	lea32_eax,		lea32_ecx, 
	lea32_edx,		lea32_ebx,
	lea32_sib,		lea32_disp32,
	lea32_esi,		lea32_edi,
	lea32_eax_disp8,	lea32_ecx_disp8, 
	lea32_edx_disp8,	lea32_ebx_disp8,
	lea32_sib_disp8,	lea32_ebp_disp8,
	lea32_esi_disp8,	lea32_edi_disp8,
	lea32_eax_disp32,	lea32_ecx_disp32, 
	lea32_edx_disp32,	lea32_ebx_disp32,
	lea32_sib_disp32,	lea32_ebp_disp32,
	lea32_esi_disp32,	lea32_edi_disp32,
};


/*
 * init table
 */
void
resolve_init(void)
{
	int i, pos;

	for (i = 0; i < 0xc0; ++i) {
		pos = ((i >> 3) & 0x18) + (i & 0x07);
		calc_ea_dst_tbl[i] = c_ea_dst_tbl[pos];
		calc_ea32_dst_tbl[i] = c_ea32_dst_tbl[pos];
		calc_lea_tbl[i] = c_lea_tbl[pos];
		calc_lea32_tbl[i] = c_lea32_tbl[pos];
	}
	for (; i < 0x100; ++i) {
		calc_ea_dst_tbl[i] = ea_nop;
		calc_ea32_dst_tbl[i] = ea32_nop;
	}
}
