/*	$Id: segments.c,v 1.1 2003/07/22 15:09:44 nonaka Exp $	*/

/*
 * Copyright (c) 2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "cpu.h"
#include "ia32.mcr"


void
load_segreg(int idx, WORD selector)
{
	SEGDESC sd;		/* segment register cache */
	DWORD sd_idx;		/* segment descriptor index */
	DWORD dt_base;		/* descriptor table base addr */
	WORD dt_limit;		/* descriptor table size limit */
	int sd_rpl;		/* segment descriptor RPL */

	if ((unsigned int)idx >= CPU_SEGREG_NUM) {
		ia32_panic("segments_load: sreg(%d)", idx);
		/*NOTREACHED*/
	}

	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* real-mode or vm86 mode */
		CPU_REGS_SREG(idx) = selector;
		CPU_STAT_SREGBASE(idx) = selector << 4;
		return;
	}

	/*
	 * protected mode
	 */
	if (idx == CPU_CS_INDEX) {
		ia32_panic("segments_load: sreg(%d)", idx);
		/*NOTREACHED*/
	}

	sd_idx = selector & ~7;
	sd_rpl = selector & 3;

	/* null segment */
	if (selector < 4) {
		switch (idx) {
		case CPU_SS_INDEX:
			EXCEPTION(GP_EXCEPTION, 0);
			/*NOTREACHED*/
			break;
		}
		/* ES, DS, FS, GS */
		CPU_REGS_SREG(idx) = selector;
		CPU_STAT_SREG_CLEAR(idx);
		return;
	}

	/* descriptor table */
	if (selector & CPU_SEGMENT_TABLE_IND) {
		/* LDT */
		if (!CPU_LDTR_DESC.valid) {
			EXCEPTION(GP_EXCEPTION, selector);
			/*NOTREACHED*/
			return;
		}
		dt_base = CPU_LDTR_BASE;
		dt_limit = CPU_LDTR_LIMIT;
	} else {
		/* GDT */
		dt_base = CPU_GDTR_BASE;
		dt_limit = CPU_GDTR_LIMIT;
	}
	if (sd_idx >= dt_limit) {
		EXCEPTION(GP_EXCEPTION, selector);
		/*NOTREACHED*/
		return;
	}
	CPU_SET_SEGDESC(sd, dt_base + sd_idx);

	if (!sd.valid) {
		EXCEPTION(GP_EXCEPTION, selector);
		/*NOTREACHED*/
		return;
	}

	switch (idx) {
	case CPU_SS_INDEX:
		if (!sd.p) {
			EXCEPTION(SS_EXCEPTION, selector);
			/*NOTREACHED*/
			return;
		}
		if ((CPU_INST_CPL != sd_rpl) ||
		    !sd.s || !sd.dc || !sd.wr ||
		    (CPU_INST_CPL != sd.dpl)) {
			EXCEPTION(GP_EXCEPTION, selector);
			/*NOTREACHED*/
			return;
		}
		CPU_INST_SS32 = sd.bd;
		CPU_REGS_SREG(idx) = selector;
		CPU_STAT_SREG(idx) = sd;
		CPU_SET_SEGDESC_POSTPART(sd);
		break;
	
	default:	/* ES, DS, FS, GS */
		if (!sd.p) {
			EXCEPTION(NP_EXCEPTION, selector);
			/*NOTREACHED*/
			return;
		}
		if (!sd.s || (sd.dc && !(sd.type & CPU_SEGDESC_TYPE_WR))) {
			EXCEPTION(GP_EXCEPTION, selector);
			/*NOTREACHED*/
			return;
		}
		if (!sd.dc || !(sd.type & CPU_SEGDESC_TYPE_EC)) {
			if ((sd_rpl > sd.dpl) || (CPU_INST_CPL > sd.dpl)) {
				EXCEPTION(GP_EXCEPTION, selector);
				/*NOTREACHED*/
				return;
			}
		}
		CPU_REGS_SREG(idx) = selector;
		CPU_STAT_SREG(idx) = sd;
		CPU_SET_SEGDESC_POSTPART(sd);
		break;
	}
}

void
load_ldtr(WORD selector)
{
	SEGDESC sd;		/* segment descriptor cache */
	DWORD sd_idx;		/* segment descriptor index */
	int sd_rpl;		/* segment descriptor RPL */

	sd_idx = selector & ~7;
	sd_rpl = selector & 3;

	/* null segment */
	if (selector < 4) {
		CPU_LDTR = selector;
		CPU_LDTR_DESC = sd;
		return;
	}

	/* descriptor table */
	if (sd_idx >= CPU_GDTR_LIMIT) {
		EXCEPTION(GP_EXCEPTION, selector);
		/*NOTREACHED*/
		return;
	}
	if (selector & CPU_SEGMENT_TABLE_IND) {
		EXCEPTION(GP_EXCEPTION, selector);
		/*NOTREACHED*/
		return;
	}
	CPU_SET_SEGDESC(sd, CPU_GDTR_BASE + sd_idx);

	if (!sd.valid || sd.s || (sd.type != 2)) {
		EXCEPTION(GP_EXCEPTION, selector);
		/*NOTREACHED*/
		return;
	}
	if (!sd.p) {
		EXCEPTION(NP_EXCEPTION, selector);
		/*NOTREACHED*/
		return;
	}
	CPU_LDTR = selector;
	CPU_LDTR_DESC = sd;
}
