/*	$Id: gtk_config.c,v 1.10 2003/06/05 18:11:23 nonaka Exp $	*/

/*
 * Copyright (c) 2002-2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "compiler.h"

#include "np2.h"
#include "pc_core.h"
#include "ini.h"
#include "xnp2.h"
#include "xmenu.h"

static GtkWidget *combo_entry1;
static GtkWidget *combo_entry2;
static GtkWidget *buffer_entry;
static int rate;

static void
ok_button_clicked(GtkButton *b, gpointer d)
{
	gchar *bufp = gtk_entry_get_text(GTK_ENTRY(buffer_entry));
	gchar *base = gtk_entry_get_text(GTK_ENTRY(combo_entry1));
	guint mul = atoi(gtk_entry_get_text(GTK_ENTRY(combo_entry2)));
	gint bufsize;

	UNUSED(b);

	if (strcmp(base, "1.9968MHz") == 0) {
		if (np2cfg.baseclock != PCBASECLOCK20) {
			INIT_RENEWAL;
			np2cfg.baseclock = PCBASECLOCK20;
		}
	} else {
		if (np2cfg.baseclock != PCBASECLOCK25) {
			INIT_RENEWAL;
			np2cfg.baseclock = PCBASECLOCK25;
		}
	}

	switch (mul) {
	case 1: case 2: case 4: case 5: case 6: case 8: case 10: case 12:
	case 16: case 20:
		if (mul != np2cfg.multiple) {
			INIT_RENEWAL;
			np2cfg.multiple = mul;
		}
	}

	switch (rate) {
	case 11025:
	case 22050:
	case 44100:
		if (rate != np2cfg.samplingrate) {
			INIT_RENEWAL;
			np2cfg.samplingrate = rate;
		}
	}

	if (bufp != NULL) {
		bufsize = atoi(bufp);
		if (bufsize < 50)
			bufsize = 50;
		else if (bufsize > 1000)
			bufsize = 1000;

		if (np2cfg.delayms != bufsize) {
			INIT_RENEWAL;
			np2cfg.delayms = bufsize;
		}
	}

	gtk_widget_destroy((GtkWidget *)d);
}

static void
config_destroy(GtkWidget *w, GtkWidget **wp)
{

	UNUSED(wp);

	install_idle_process();
	gtk_widget_destroy(w);
}

static void
radiobutton_clicked(GtkButton *b, gpointer d)
{

	UNUSED(b);

	rate = (gint)d;
}

static void
clock_changed(GtkEditable *e, gpointer d)
{
	GtkWidget *w = (GtkWidget *)d;
	gchar *base = gtk_entry_get_text(GTK_ENTRY(combo_entry1));
	guint mul = atoi(gtk_entry_get_text(GTK_ENTRY(combo_entry2)));
	gchar buf[80];
	gint clock;

	UNUSED(e);

	if (strcmp(base, "1.9968MHz") == 0) {
		clock = PCBASECLOCK20 * mul;
	} else {
		clock = PCBASECLOCK25 * mul;
	}
	g_snprintf(buf, sizeof(buf), "%2d.%4dMHz",
	    clock / 1000000U, (clock / 1000) % 1000);
	gtk_label_set_text(GTK_LABEL(w), buf);
}

void
create_configure_dialog(void)
{
	GtkWidget *config_window;
	GtkWidget *vbox1, *vbox2, *vbox3;
	GtkWidget *hbox1, *hbox2, *hbox3, *hbox4;
	GtkWidget *alignment1, *alignment2;
	GtkWidget *radiobutton[3];
	GtkWidget *cpuframe;
	GtkWidget *baseclock_combo;
	GtkWidget *rate_combo;
	GtkWidget *times_label;
	GtkWidget *realclock_label;
	GtkWidget *confirm;
	GtkWidget *ok_button;
	GtkWidget *cancel_button;
	GtkWidget *sound_frame;
	GtkWidget *rate_label;
	GtkWidget *buffer_label;
	GtkWidget *ms_label;
	GList *baseclock_combo_items = NULL;
	GList *rate_combo_items = NULL;
	GSList *hbox4_group = NULL;
	gchar buf[8];
	gint i;

	uninstall_idle_process();

	config_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data(GTK_OBJECT(config_window), "config_window",
	    config_window);
	gtk_widget_set_usize(config_window, 320, 200);
	gtk_window_set_title(GTK_WINDOW(config_window), "Configure");
	gtk_window_set_position(GTK_WINDOW(config_window), GTK_WIN_POS_CENTER);
	gtk_window_set_modal(GTK_WINDOW(config_window), TRUE);
	gtk_window_set_policy(GTK_WINDOW(config_window), TRUE, TRUE, TRUE);

	gtk_signal_connect(GTK_OBJECT(config_window), "destroy",
	    GTK_SIGNAL_FUNC(config_destroy), NULL);

	vbox1 = gtk_vbox_new(FALSE, 0);
	gtk_widget_ref(vbox1);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "vbox1", vbox1,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(vbox1);
	gtk_container_add(GTK_CONTAINER(config_window), vbox1);

	hbox1 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox1);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "hbox1", hbox1,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox1);
	gtk_box_pack_start(GTK_BOX(vbox1), hbox1, TRUE, TRUE, 0);

	alignment1 = gtk_alignment_new(0.5, 0.5, 1, 1);
	gtk_widget_ref(alignment1);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "alignment1",
	    alignment1, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(alignment1);
	gtk_box_pack_start(GTK_BOX(hbox1), alignment1, TRUE, TRUE, 0);

	cpuframe = gtk_frame_new("CPU");
	gtk_widget_ref(cpuframe);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "cpuframe",
	    cpuframe, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(cpuframe);
	gtk_container_add(GTK_CONTAINER(alignment1), cpuframe);
	gtk_container_set_border_width(GTK_CONTAINER(cpuframe), 5);

	vbox2 = gtk_vbox_new(FALSE, 0);
	gtk_widget_ref(vbox2);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "vbox2", vbox2,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(vbox2);
	gtk_container_add(GTK_CONTAINER(cpuframe), vbox2);

	hbox2 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox2);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "hbox2", hbox2,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox2);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox2, TRUE, TRUE, 0);

	baseclock_combo = gtk_combo_new();
	gtk_widget_ref(baseclock_combo);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "baseclock_combo",
	    baseclock_combo, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(baseclock_combo);
	gtk_box_pack_start(GTK_BOX(hbox2), baseclock_combo, TRUE, FALSE, 0);
	gtk_widget_set_usize(baseclock_combo, 96, -2);
	gtk_combo_set_value_in_list(GTK_COMBO(baseclock_combo), TRUE, TRUE);
	gtk_combo_set_use_arrows_always(GTK_COMBO(baseclock_combo), TRUE);
	baseclock_combo_items = g_list_append(baseclock_combo_items,
	    (gpointer)"1.9968MHz");
	baseclock_combo_items = g_list_append(baseclock_combo_items,
	    (gpointer)"2.4576MHz");
	gtk_combo_set_popdown_strings(GTK_COMBO(baseclock_combo),
	    baseclock_combo_items);
	g_list_free(baseclock_combo_items);

	combo_entry1 = GTK_COMBO(baseclock_combo)->entry;
	gtk_widget_ref(combo_entry1);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "combo_entry1",
	    combo_entry1, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(combo_entry1);
	gtk_entry_set_editable(GTK_ENTRY(combo_entry1), FALSE);
	if (np2cfg.baseclock == PCBASECLOCK20)
		gtk_entry_set_text(GTK_ENTRY(combo_entry1), "1.9968MHz");
	else
		gtk_entry_set_text(GTK_ENTRY(combo_entry1), "2.4576MHz");

	times_label = gtk_label_new("x");
	gtk_widget_ref(times_label);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "times_label",
	    times_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(times_label);
	gtk_box_pack_start(GTK_BOX(hbox2), times_label, TRUE, FALSE, 0);
	gtk_misc_set_padding(GTK_MISC(times_label), 5, 0);

	rate_combo = gtk_combo_new();
	gtk_widget_ref(rate_combo);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "rate_combo",
	    rate_combo, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(rate_combo);
	gtk_box_pack_start(GTK_BOX(hbox2), rate_combo, TRUE, FALSE, 0);
	gtk_widget_set_usize(rate_combo, 48, -2);
	gtk_combo_set_value_in_list(GTK_COMBO(rate_combo), TRUE, TRUE);
	gtk_combo_set_use_arrows_always(GTK_COMBO(rate_combo), TRUE);
	rate_combo_items = g_list_append(rate_combo_items, (gpointer)"1");
	rate_combo_items = g_list_append(rate_combo_items, (gpointer)"2");
	rate_combo_items = g_list_append(rate_combo_items, (gpointer)"4");
	rate_combo_items = g_list_append(rate_combo_items, (gpointer)"5");
	rate_combo_items = g_list_append(rate_combo_items, (gpointer)"6");
	rate_combo_items = g_list_append(rate_combo_items, (gpointer)"8");
	rate_combo_items = g_list_append(rate_combo_items, (gpointer)"10");
	rate_combo_items = g_list_append(rate_combo_items, (gpointer)"12");
	rate_combo_items = g_list_append(rate_combo_items, (gpointer)"16");
	rate_combo_items = g_list_append(rate_combo_items, (gpointer)"20");
	gtk_combo_set_popdown_strings(GTK_COMBO(rate_combo), rate_combo_items);
	g_list_free(rate_combo_items);

	combo_entry2 = GTK_COMBO(rate_combo)->entry;
	gtk_widget_ref(combo_entry2);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "combo_entry2",
	    combo_entry2, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(combo_entry2);
	gtk_entry_set_editable(GTK_ENTRY(combo_entry2), FALSE);
	switch (np2cfg.multiple) {
	case 1: case 2: case 4: case 5: case 6: case 8: case 10: case 12:
	case 16: case 20:
		g_snprintf(buf, sizeof(buf), "%d", np2cfg.multiple);
		gtk_entry_set_text(GTK_ENTRY(combo_entry2), buf);
		break;

	default:
		gtk_entry_set_text(GTK_ENTRY(combo_entry2), "1");
		break;
	}

	realclock_label = gtk_label_new("MHz");
	gtk_widget_ref(realclock_label);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "realclock_label",
	    realclock_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(realclock_label);
	gtk_box_pack_start(GTK_BOX(vbox2), realclock_label, FALSE, FALSE, 0);
	gtk_label_set_justify(GTK_LABEL(realclock_label), GTK_JUSTIFY_RIGHT);
	gtk_misc_set_alignment(GTK_MISC(realclock_label), 1, 0);
	gtk_signal_connect(GTK_OBJECT(combo_entry1), "changed",
	  GTK_SIGNAL_FUNC(clock_changed), (gpointer)realclock_label);
	gtk_signal_connect(GTK_OBJECT(combo_entry2), "changed",
	  GTK_SIGNAL_FUNC(clock_changed), (gpointer)realclock_label);
	clock_changed(NULL, realclock_label);

	confirm = gtk_vbutton_box_new();
	gtk_widget_ref(confirm);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "confirm", confirm,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(confirm);
	gtk_box_pack_start(GTK_BOX(hbox1), confirm, TRUE, TRUE, 0);
	gtk_button_box_set_layout(GTK_BUTTON_BOX(confirm), GTK_BUTTONBOX_END);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(confirm), 0);

	sound_frame = gtk_frame_new("Sound");
	gtk_widget_ref(sound_frame);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "sound_frame",
	    sound_frame, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(sound_frame);
	gtk_box_pack_start(GTK_BOX(vbox1), sound_frame, TRUE, TRUE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(sound_frame), 5);

	vbox3 = gtk_vbox_new(FALSE, 0);
	gtk_widget_ref(vbox3);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "vbox3", vbox3,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(vbox3);
	gtk_container_add(GTK_CONTAINER(sound_frame), vbox3);

	hbox4 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox4);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "hbox4", hbox4,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox4);
	gtk_box_pack_start(GTK_BOX(vbox3), hbox4, TRUE, TRUE, 0);

	rate_label = gtk_label_new("Sampling Rate");
	gtk_widget_ref(rate_label);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "rate_label",
	    rate_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(rate_label);
	gtk_box_pack_start(GTK_BOX(hbox4), rate_label, FALSE, TRUE, 3);
	gtk_widget_set_usize(rate_label, 96, -2);

	radiobutton[0] = gtk_radio_button_new_with_label(hbox4_group, "11KHz");
	hbox4_group = gtk_radio_button_group(GTK_RADIO_BUTTON(radiobutton[0]));
	gtk_widget_ref(radiobutton[0]);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "radiobutton0",
	    radiobutton[0], (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(radiobutton[0]);
	gtk_box_pack_start(GTK_BOX(hbox4), radiobutton[0], FALSE, FALSE, 0);
	GTK_WIDGET_UNSET_FLAGS(radiobutton[0], GTK_CAN_FOCUS);
	gtk_signal_connect(GTK_OBJECT(radiobutton[0]), "clicked",
	    GTK_SIGNAL_FUNC(radiobutton_clicked), (gpointer)11025);

	radiobutton[1] = gtk_radio_button_new_with_label(hbox4_group, "22KHz");
	hbox4_group = gtk_radio_button_group(GTK_RADIO_BUTTON(radiobutton[1]));
	gtk_widget_ref(radiobutton[1]);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "radiobutton1",
	    radiobutton[1], (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(radiobutton[1]);
	gtk_box_pack_start(GTK_BOX(hbox4), radiobutton[1], FALSE, FALSE, 0);
	GTK_WIDGET_UNSET_FLAGS(radiobutton[1], GTK_CAN_FOCUS);
	gtk_signal_connect(GTK_OBJECT(radiobutton[1]), "clicked",
	    GTK_SIGNAL_FUNC(radiobutton_clicked), (gpointer)22050);

	radiobutton[2] = gtk_radio_button_new_with_label(hbox4_group, "44KHz");
	hbox4_group = gtk_radio_button_group(GTK_RADIO_BUTTON(radiobutton[2]));
	gtk_widget_ref(radiobutton[2]);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "radiobutton2",
	    radiobutton[2], (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(radiobutton[2]);
	gtk_box_pack_start(GTK_BOX(hbox4), radiobutton[2], FALSE, FALSE, 0);
	GTK_WIDGET_UNSET_FLAGS(radiobutton[2], GTK_CAN_FOCUS);
	gtk_signal_connect(GTK_OBJECT(radiobutton[2]), "clicked",
	    GTK_SIGNAL_FUNC(radiobutton_clicked), (gpointer)44100);

	if (np2cfg.samplingrate < AVE(11025, 22050)) {
		i = 0;
	} else if (np2cfg.samplingrate < AVE(22050, 44100)) {
		i = 1;
	} else {
		i = 2;
	}
	gtk_signal_emit_by_name(GTK_OBJECT(radiobutton[i]), "clicked");

	alignment2 = gtk_alignment_new(0.5, 0.5, 1, 1);
	gtk_widget_ref(alignment2);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "alignment2",
	    alignment2, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(alignment2);
	gtk_box_pack_start(GTK_BOX(vbox3), alignment2, TRUE, TRUE, 0);

	hbox3 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox3);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "hbox3", hbox3,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox3);
	gtk_container_add(GTK_CONTAINER(alignment2), hbox3);

	buffer_label = gtk_label_new("Buffer");
	gtk_widget_ref(buffer_label);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "buffer_label",
	    buffer_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(buffer_label);
	gtk_box_pack_start(GTK_BOX(hbox3), buffer_label, FALSE, FALSE, 0);
	gtk_widget_set_usize(buffer_label, 96, -2);

	buffer_entry = gtk_entry_new();
	gtk_widget_ref(buffer_entry);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "buffer_entry",
	    buffer_entry, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(buffer_entry);
	gtk_box_pack_start(GTK_BOX(hbox3), buffer_entry, FALSE, FALSE, 0);
	gtk_widget_set_usize(buffer_entry, 48, -2);
	if (np2cfg.delayms >= 50 && np2cfg.delayms <= 1000) {
		g_snprintf(buf, sizeof(buf), "%d", np2cfg.delayms);
		gtk_entry_set_text(GTK_ENTRY(buffer_entry), buf);
	} else
		gtk_entry_set_text(GTK_ENTRY(buffer_entry), "800");

	ms_label = gtk_label_new(" ms");
	gtk_widget_ref(ms_label);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "ms_label",
	    ms_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(ms_label);
	gtk_box_pack_start(GTK_BOX(hbox3), ms_label, FALSE, FALSE, 0);
	gtk_label_set_justify(GTK_LABEL(ms_label), GTK_JUSTIFY_RIGHT);

	ok_button = gtk_button_new_with_label("OK");
	gtk_widget_ref(ok_button);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "ok_button",
	    ok_button, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(ok_button);
	gtk_container_add(GTK_CONTAINER(confirm), ok_button);
	GTK_WIDGET_SET_FLAGS(ok_button, GTK_CAN_DEFAULT);
	gtk_signal_connect(GTK_OBJECT(ok_button), "clicked",
	    GTK_SIGNAL_FUNC(ok_button_clicked), (gpointer)config_window);

	cancel_button = gtk_button_new_with_label("Cancel");
	gtk_widget_ref(cancel_button);
	gtk_object_set_data_full(GTK_OBJECT(config_window), "cancel_button",
	    cancel_button, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(cancel_button);
	gtk_container_add(GTK_CONTAINER(confirm), cancel_button);
	GTK_WIDGET_SET_FLAGS(cancel_button, GTK_CAN_DEFAULT);
	gtk_signal_connect_object(GTK_OBJECT(cancel_button), "clicked",
	    GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(config_window));

	gtk_widget_grab_default(ok_button);

	gtk_widget_show_all(config_window);
}
