/*	$Id: gtkfont.c,v 1.9 2003/06/05 18:11:23 nonaka Exp $	*/

/*
 * Copyright (c) 2002-2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "compiler.h"

#include "np2.h"
#include "dosio.h"
#include "font.h"
#include "parts.h"
#include "xnp2.h"

#ifdef NP2_BIGENDIAN
static unsigned short
LE_short(unsigned short v)
{

	return ((((v) & 0xff) << 8) | (((v) >> 8) & 0xff));
}

static unsigned long
LE_long(unsigned long v)
{

	return ((((v) & 0xff) << 24)
	      | (((v) & 0xff00) << 8)
	      | (((v) >> 8) & 0xff00)
	      | (((v) >> 24) & 0xff));
}

#define	BE_short(v)	(v)
#define	BE_long(v)	(v)

#else	/* !NP2_BIGENDIAN */

#define	LE_short(v)	(v)
#define	LE_long(v)	(v)

static WORD
BE_short(WORD x)
{

	return ((((x) & 0xff00) >> 8)
		| (((x) & 0xff) << 8));
}

#if 0
static DWORD
BE_long(DWORD x)
{

	return ((((x) & 0xff000000) >> 24)
		| (((x) & 0x00ff0000) >> 8)
		| (((x) & 0x0000ff00) << 8)
		| (((x) & 0x000000ff) << 24));
}
#endif

#endif	/* NP2_BIGENDIAN */

static WORD
jis2euc(WORD jis)
{

	return ((jis + 0x80) & 0xff) + ((jis + 0x8000) & 0xff00);
}

#if 0
static WORD
euc2jis(WORD euc)
{

#if 0
	if ((euc & 0xff00) == 0x8e00)
		hankaku2zenkaku(euc & 0xff);
	else
		return ((euc - 0x80) & 0xff) + ((euc - 0x8000) & 0xff00);
#else
	return ((euc - 0x80) & 0xff) + ((euc - 0x8000) & 0xff00);
#endif
}

#if 0
static WORD
hankaku2zen(int code)
{
	static WORD cnv[64] = {
		0x2121, 0x2123, 0x2156, 0x2157, 0x2122, 0x2126, 0x2572, 0x2521,
       		0x2523, 0x2525, 0x2527, 0x2529, 0x2563, 0x2565, 0x2567, 0x2543,
		0x213c, 0x2522, 0x2524, 0x2526, 0x2528, 0x252a, 0x252b, 0x252d,
		0x252f, 0x2531, 0x2533, 0x2535, 0x2537, 0x2539, 0x253b, 0x253d,
		0x253f, 0x2541, 0x2544, 0x2546, 0x2548, 0x254a, 0x254b, 0x254c,
		0x254d, 0x254e, 0x254f, 0x2552, 0x2555, 0x2558, 0x255b, 0x255e,
		0x255f, 0x2560, 0x2561, 0x2562, 0x2564, 0x2566, 0x2568, 0x2569,
		0x256a, 0x256b, 0x256c, 0x256d, 0x256f, 0x2573, 0x212b, 0x212c
	};

	code -= 0xa0;
	if (code < 0 || code > sizeof(cnv) / sizeof(cnv[0]))
		return 0;
	return cnv[code];
}
#endif
#endif

static GdkPixmap *font_pixmap = NULL;
static unsigned long black_pixel;

static int deltable[] = {
	0x0f, 0x5f, 0,
	0x01, 0x10, 0x1a, 0x21, 0x3b, 0x41, 0x5b, 0x5f, 0,
	0x54, 0x5f, 0,
	0x57, 0x5f, 0,
	0x19, 0x21, 0x39, 0x5f, 0,
	0x22, 0x31, 0x52, 0x5f, 0,
	0x01, 0x5f, 0,
	0x01, 0x5f, 0,
	0x01, 0x5f, 0,
	0x01, 0x5f, 0,
	0x01, 0x5f, 0,
	0x1f, 0x20, 0x37, 0x3f, 0x5d, 0x5f, 0
};

static BYTE jis90_83[] = "y}|A˚融增wM";
static BYTE jis90_90[] = "ahЌz{ǐxGٓvOU";

static BYTE jis83_78[] = "Ėy";
static BYTE jis83_83[] = "";	//  PC-98ŊJĂ邩ǂ߂c

#ifndef	FONTFACE
#define	FONTFACE "-sony-fixed-medium-r-normal--16-*-*-*-*-*-jisx0201.1976-0,"\
		"-jis-fixed-medium-r-normal--16-*-*-*-*-*-jisx0208.1983-0"
#endif

static BYTE necfont[] = {
/* o1 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x41, 0x04, 
0x43, 0x04, 0x81, 0x02, 0x81, 0x02, 0x81, 0x02, 
0x81, 0x02, 0x81, 0x02, 0x41, 0x04, 0x43, 0x84, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o2 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x43, 0x84, 
0x44, 0x44, 0x88, 0x22, 0x88, 0x22, 0x80, 0x42, 
0x81, 0x82, 0x82, 0x02, 0x44, 0x04, 0x4f, 0xe4, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o3 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x43, 0x84, 
0x44, 0x44, 0x88, 0x22, 0x80, 0x22, 0x81, 0xc2, 
0x80, 0x22, 0x88, 0x22, 0x44, 0x44, 0x43, 0x84, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o4 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x40, 0x44, 
0x40, 0xc4, 0x81, 0x42, 0x82, 0x42, 0x82, 0x42, 
0x84, 0x42, 0x8f, 0xe2, 0x40, 0x44, 0x40, 0x44, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o5 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x44, 0x04, 
0x47, 0xe4, 0x84, 0x02, 0x84, 0x02, 0x87, 0xc2, 
0x80, 0x22, 0x80, 0x22, 0x44, 0x24, 0x43, 0xc4, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o6 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x43, 0x84, 
0x44, 0x44, 0x88, 0x22, 0x88, 0x02, 0x8b, 0xc2, 
0x8c, 0x22, 0x88, 0x22, 0x48, 0x24, 0x44, 0x44, 
0x23, 0x88, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o7 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x4f, 0xe4, 
0x48, 0x24, 0x88, 0x22, 0x80, 0x42, 0x80, 0x42, 
0x80, 0x82, 0x80, 0x82, 0x41, 0x04, 0x41, 0x04, 
0x21, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o8 */
0x07, 0xc0, 0x18, 0x30, 0x23, 0x88, 0x44, 0x44, 
0x48, 0x24, 0x88, 0x22, 0x84, 0x42, 0x83, 0x82, 
0x84, 0x42, 0x88, 0x22, 0x48, 0x24, 0x44, 0x44, 
0x23, 0x88, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o9 */
0x07, 0xc0, 0x18, 0x30, 0x23, 0x88, 0x44, 0x44,
0x48, 0x24, 0x88, 0x22, 0x88, 0x22, 0x84, 0x62,
0x83, 0xa2, 0x80, 0x22, 0x48, 0x24, 0x44, 0x44,
0x23, 0x88, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00,

/* o10 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x44, 0xc4, 
0x4d, 0x24, 0x85, 0x22, 0x85, 0x22, 0x85, 0x22, 
0x85, 0x22, 0x85, 0x22, 0x45, 0x24, 0x44, 0xc4, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o11 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x44, 0x44, 
0x4c, 0xc4, 0x84, 0x42, 0x84, 0x42, 0x84, 0x42, 
0x84, 0x42, 0x84, 0x42, 0x44, 0x44, 0x44, 0x44, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o12 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x44, 0xc4, 
0x4d, 0x24, 0x85, 0x22, 0x84, 0x22, 0x84, 0x42, 
0x84, 0x42, 0x84, 0x82, 0x45, 0x04, 0x45, 0xe4, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o13 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x44, 0xc4, 
0x4d, 0x24, 0x84, 0x22, 0x84, 0x22, 0x84, 0x42, 
0x84, 0x22, 0x84, 0x22, 0x45, 0x24, 0x44, 0xc4, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o14 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x44, 0x24, 
0x4c, 0x64, 0x84, 0xa2, 0x84, 0xa2, 0x84, 0xa2, 
0x85, 0x22, 0x85, 0xf2, 0x44, 0x24, 0x44, 0x24, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o15 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x45, 0x04, 
0x4d, 0xf4, 0x85, 0x02, 0x85, 0x02, 0x85, 0xe2, 
0x84, 0x12, 0x84, 0x12, 0x45, 0x14, 0x44, 0xe4, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o16 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x44, 0xe4, 
0x4d, 0x14, 0x85, 0x12, 0x85, 0x02, 0x85, 0xe2, 
0x85, 0x12, 0x85, 0x12, 0x45, 0x14, 0x44, 0xe4, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o17 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x45, 0xf4, 
0x4d, 0x14, 0x85, 0x12, 0x84, 0x22, 0x84, 0x22, 
0x84, 0x22, 0x84, 0x42, 0x44, 0x44, 0x44, 0x44, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o18 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x44, 0xe4, 
0x4d, 0x14, 0x85, 0x12, 0x84, 0xa2, 0x84, 0x42, 
0x84, 0xa2, 0x85, 0x12, 0x45, 0x14, 0x44, 0xe4, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o19 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x44, 0xe4, 
0x4d, 0x14, 0x85, 0x12, 0x85, 0x32, 0x84, 0xd2, 
0x84, 0x12, 0x84, 0x12, 0x45, 0x14, 0x44, 0xe4, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* o20 */
0x07, 0xc0, 0x18, 0x30, 0x20, 0x08, 0x4c, 0x64, 
0x52, 0x94, 0x92, 0x92, 0x82, 0x92, 0x84, 0x92, 
0x84, 0x92, 0x84, 0x92, 0x48, 0x94, 0x5e, 0x64, 
0x20, 0x08, 0x18, 0x30, 0x07, 0xc0, 0x00, 0x00, 

/* I */
0x00, 0x00, 0x03, 0x80, 0x01, 0x00, 0x01, 0x00, 
0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 
0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 
0x01, 0x00, 0x01, 0x00, 0x03, 0x80, 0x00, 0x00, 

/* II */
0x00, 0x00, 0x07, 0xc0, 0x02, 0x80, 0x02, 0x80, 
0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 
0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 
0x02, 0x80, 0x02, 0x80, 0x07, 0xc0, 0x00, 0x00, 

/* III */
0x00, 0x00, 0x0f, 0xe0, 0x05, 0x40, 0x05, 0x40, 
0x05, 0x40, 0x05, 0x40, 0x05, 0x40, 0x05, 0x40, 
0x05, 0x40, 0x05, 0x40, 0x05, 0x40, 0x05, 0x40, 
0x05, 0x40, 0x05, 0x40, 0x0f, 0xe0, 0x00, 0x00, 

/* IV */
0x00, 0x00, 0x3f, 0xf8, 0x14, 0x10, 0x14, 0x10, 
0x14, 0x10, 0x12, 0x20, 0x12, 0x20, 0x12, 0x20, 
0x11, 0x40, 0x11, 0x40, 0x11, 0x40, 0x10, 0x80, 
0x10, 0x80, 0x10, 0x80, 0x38, 0x80, 0x00, 0x00, 

/* V */
0x00, 0x00, 0x3f, 0xf8, 0x10, 0x10, 0x10, 0x10, 
0x10, 0x10, 0x08, 0x20, 0x08, 0x20, 0x04, 0x40, 
0x04, 0x40, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 
0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x00, 

/* VI */
0x00, 0x00, 0x3f, 0xf8, 0x10, 0x50, 0x10, 0x50, 
0x10, 0x50, 0x08, 0x90, 0x08, 0x90, 0x08, 0x90, 
0x05, 0x10, 0x05, 0x10, 0x05, 0x10, 0x02, 0x10, 
0x02, 0x10, 0x02, 0x10, 0x02, 0x38, 0x00, 0x00, 

/* VII */
0x00, 0x00, 0x7f, 0xfc, 0x20, 0xa8, 0x20, 0xa8, 
0x20, 0xa8, 0x11, 0x28, 0x11, 0x28, 0x11, 0x28, 
0x0a, 0x28, 0x0a, 0x28, 0x0a, 0x28, 0x04, 0x28, 
0x04, 0x28, 0x04, 0x28, 0x04, 0x7c, 0x00, 0x00, 

/* VIII */
0x00, 0x00, 0x7f, 0xff, 0x20, 0xaa, 0x20, 0xaa, 
0x20, 0xaa, 0x11, 0x2a, 0x11, 0x2a, 0x11, 0x2a, 
0x0a, 0x2a, 0x0a, 0x2a, 0x0a, 0x2a, 0x04, 0x2a, 
0x04, 0x2a, 0x04, 0x2a, 0x04, 0x7f, 0x00, 0x00, 

/* IX */
0x00, 0x00, 0x3f, 0xf8, 0x10, 0x50, 0x10, 0x50, 
0x08, 0x90, 0x08, 0x90, 0x05, 0x10, 0x02, 0x10, 
0x02, 0x10, 0x05, 0x10, 0x08, 0x90, 0x08, 0x90, 
0x10, 0x50, 0x10, 0x50, 0x3f, 0xf8, 0x00, 0x00, 

/* X */
0x00, 0x00, 0x3f, 0xf8, 0x10, 0x10, 0x10, 0x10, 
0x08, 0x20, 0x04, 0x40, 0x02, 0x80, 0x01, 0x00, 
0x01, 0x00, 0x02, 0x80, 0x04, 0x40, 0x08, 0x20, 
0x10, 0x10, 0x10, 0x10, 0x3f, 0xf8, 0x00, 0x00, 
};

static int
getmonobmpsize(int x, int y)
{

        return sizeof(BITMAPINFOHEADER) + 4 * 2 + (((x + 31) / 8) & ~3) * y;
}

static void
cnvjis78_83(char *str)
{
	int i;

	while ((*str) && (*(str+1))) {
		// jis78jis83
		for (i = 0; jis83_78[i]; i += 2) {
			if (*(WORD *)str == *(WORD *)(&jis83_78[i])) {
				*(WORD *)str = *(WORD *)(&jis83_83[i]);
				break;
			}
		}
		str += 2;
	}
}

static void
cnvjis83_90(char *str)
{
	int i;

	while ((*str) && (*(str+1))) {
		// jis83jis90
		for (i = 0; jis90_83[i]; i += 2) {
			if (*(WORD *)str == *(WORD *)(&jis90_83[i])) {
				*(WORD *)str = *(WORD *)(&jis90_90[i]);
				break;
			}
			if (*(WORD *)str == *(WORD *)(&jis90_90[i])) {
				*(WORD *)str = *(WORD *)(&jis90_83[i]);
				break;
			}
		}
		str += 2;
	}
}

static void
setmonobmphead(BITMAPINFO *bi, int x, int y)
{
	BYTE i;

	ZeroMemory(&bi->bmiHeader, sizeof(BITMAPINFOHEADER));
	bi->bmiHeader.biSize = LE_long(sizeof(BITMAPINFOHEADER));
	bi->bmiHeader.biWidth = LE_long(x);
	bi->bmiHeader.biHeight = LE_long(y);
	bi->bmiHeader.biPlanes = LE_short(1);
	bi->bmiHeader.biBitCount = LE_short(1);
	bi->bmiHeader.biCompression = LE_long(BI_RGB);
	bi->bmiHeader.biSizeImage = LE_long((((x + 31) / 8) & ~3) * y);
	bi->bmiHeader.biXPelsPerMeter = LE_long(0);
	bi->bmiHeader.biYPelsPerMeter = LE_long(0);
	bi->bmiHeader.biClrUsed = LE_long(2);
	bi->bmiHeader.biClrImportant = LE_long(2);

	for (i = 0; i < 2; i++) {
		bi->bmiColors[i].rgbRed = (i ^ 1) - 1;
		bi->bmiColors[i].rgbGreen = (i ^ 1) - 1;
		bi->bmiColors[i].rgbBlue = (i ^ 1) - 1;
		bi->bmiColors[i].rgbReserved = 1;
	}
}

static BOOL
getfont_sub(BYTE *dest, char *str, GdkFont *fontset, int fntx, int fnty,
    int bmpx, int bmpy) {
	GdkImage *font_image;
	unsigned long pixel;
	int len;
	int x, y;
	unsigned char mask;
	unsigned char *p;

	UNUSED(bmpx);
	UNUSED(bmpy);

	len = strlen(str);
	gdk_draw_rectangle(font_pixmap, window->style->white_gc, TRUE,
	    0, 0, 256 * 8, 16);
	gdk_draw_text(font_pixmap, fontset, window->style->black_gc,
	    0, 14, str, len);
	font_image = gdk_image_get(font_pixmap, 0, 0, len * fntx, fnty);
	if (font_image == NULL) {
		fprintf(stderr, "Can't get ximage.\n");
		return FALSE;
	}

	for (y = 0; y < fnty ; y++) {
		p = dest + (fnty - 1 - y) * 256;
		mask = 0x80;
		for (x = 0; x < len * fntx; x++) {
			pixel = gdk_image_get_pixel(font_image, x, y);
			if (pixel == black_pixel)
				*p |= mask;
			else
				*p &= ~mask;

			if ((x % 8) == 7) {
				mask = 0x80;
				p++;
			} else
				mask >>= 1;
		}
	}

	gdk_image_destroy(font_image);

	return TRUE;
}

static BOOL
getfont(BYTE *dest, GdkFont *fontset, int base, int size)
{
	BYTE textbuf[256];
	int *p;
	int i;
	WORD code;

	if (base == 0) {
		for (i = 0; i < 0x5f; i++)
			textbuf[i] = (BYTE)(i + 0x20);
		textbuf[0x5f] = ' ';
		textbuf[0x60] = ' ';
		for (i = 0x01; i < 0x40; i++) {
			textbuf[i * 2 + 0x60 - 1] = 0x8e;
			textbuf[i * 2 + 1 + 0x60 - 1] = (BYTE)(i + 0xa0);
		}
		textbuf[0xe0] = '\0';
	} else {
		*(WORD *)textbuf = jis2euc(0x2020);
		for (i = 1; i < 0x5f; i++) {
			code = jis2euc((base << 8) + i + 0x20);
			((WORD *)textbuf)[i] = BE_short(code);
		}
		textbuf[i * 2] = '\0';
		switch (base) {
		case 0x22:
		case 0x23:
		case 0x24:
		case 0x25:
		case 0x26:
		case 0x27:
		case 0x28:
		case 0x29:
		case 0x2a:
		case 0x2b:
		case 0x2c:
		case 0x2d:
			p = deltable;
			while (base > 0x22) {
				while (*p++)
					continue;
				base--;
			}
			while (*p) {
				i = *p++;
				while (i < *p) {
					code = jis2euc(0x2020);
					((WORD *)textbuf)[i++] = BE_short(code);
				}
				p++;
			}
			break;

		case 0x4f:
			for (i = 0x34; i < 0x5f; i++) {
				code = jis2euc(0x2020);
				((WORD *)textbuf)[i] = BE_short(code);
			}
			break;

		case 0x7c:
			for (i = 0x4f; i < 0x51; i++) {
				code = jis2euc(0x2020);
				((WORD *)textbuf)[i] = BE_short(code);
			}
			break;

		case 0x2e:
		case 0x2f:
		case 0x74:
		case 0x75:
		case 0x76:
		case 0x77:
		case 0x78:
		case 0x7d:
		case 0x7e:
		case 0x7f:
			for (i = 0x01; i < 0x5f; i++) {
				code = jis2euc(0x2020);
				((WORD *)textbuf)[i] = BE_short(code);
			}
			break;
		}
		cnvjis78_83(textbuf);
		cnvjis83_90(textbuf);
	}

	return getfont_sub(dest, textbuf, fontset, size, size*2, size*256, size*2);
}

static void
patchank2bmp(BYTE *dst, WORD *src)
{
	int i, j;
	WORD v;

	for (i = 0; i < 0x20; i++) {
		for (j = 0; j < 8; j++) {
			v = ~(*src++);

			dst -= 256;
			*dst = v & 0xff;

			dst -= 256;
			*dst = (v >> 8) & 0xff;
		}
		dst += 256 * 16 + 1;
	}
}

// 2x21-2x7e
static void
patchextank2bmp(BYTE *dst, WORD *src)
{
	int size;
	WORD v;

	size = 0x5e * 8;
	while (size--) {
		v = ~(*src++);

		dst -= 256;
		*dst = v & 0xff;

		dst -= 256;
		*dst = (v >> 8) & 0xff;
	}
}

// 2c24-2c6f
static void
patchextfnt2bmp(BYTE *dst, WORD *src)
{
	int i;

	dst -= 3*16*256;
	for (i = 0; i < 0x4c * 16; i++) {
		dst -= 256;
		*(WORD *)dst = ~(*src++);
	}
}

static void
patchnecfnt2bmp(BYTE *dst, BYTE *src)
{
	int i, j;

	for (i = 0; i < (int)sizeof(necfont) / 2; i++) {
		for (j = 0; j < 2; j++) {
			*dst++ = ~(*src++);
		}
		dst -= 256 + 2;
	}
}

void
np2os_makefontbmp(char *filename)
{
	BYTE fnt[256 * 16];
	GdkColor color;
	GdkFont *fontset;
	GdkColormap *colormap;
	BYTE *work;
	BYTE *p, *q;
	FILEH *fh;
	int size;
	int offbit;
	int jish;
	int i, j;

	colormap = gtk_widget_get_colormap(GTK_WIDGET(window));
	color.red = color.green = color.blue = 0;
	if (gdk_color_alloc(colormap, &color) == 0) {
		black_pixel = 0;
	} else {
		black_pixel = color.pixel;
	}

	fontset = gdk_fontset_load(FONTFACE);
	if (fontset == NULL)
		return;
	font_pixmap = gdk_pixmap_new(window->window, 256 * 8, 16, -1);

	size = sizeof(BITMAPFILEHEADER) + getmonobmpsize(2048, 2048);
	offbit = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + 4 * 2;

	work = malloc(size);
	if (work == NULL) {
		fprintf(stderr, "Can't alloc memory (font mitmap)\n");
		gdk_pixmap_unref(font_pixmap);
		gdk_font_unref(fontset);
		return;
	}

	memset(work, 0, sizeof(BITMAPFILEHEADER));
	((BITMAPFILEHEADER *)work)->bfType = LE_short(0x4d42);
	((BITMAPFILEHEADER *)work)->bfSize = LE_long(size);
	((BITMAPFILEHEADER *)work)->bfOffBits = LE_long(offbit);
	setmonobmphead((BITMAPINFO *)(work + sizeof(BITMAPFILEHEADER)),
	    2048, 2048);
	memset(work + offbit, 0xff, 256 * 2048);

	memset(fnt, 0, sizeof(fnt));
	getfont(fnt, fontset, 0, 8);
	p = fnt;
	q = work + offbit + 256 * 16 * 0x7f;
	for (i = 0; i < 16; i++) {
		for (j = 0x00; j < 0x5f; j++)
			q[j+0x20] = ~p[j];
		for (j = 0x01; j < 0x40; j++)
			q[j + 0xa0] = ~p[j + 0x60];
		p += 256;
		q += 256;
	}
	patchank2bmp(q + 0x00, fontank16 + 0x00 * 8);
	patchank2bmp(q + 0x80, fontank16 + 0x20 * 8);
	patchank2bmp(q + 0xe0, fontank16 + 0x40 * 8);

	for (jish = 0x01; jish < 0x60; jish++) {
		memset(fnt, 0, sizeof(fnt));
		getfont(fnt, fontset, jish + 0x20, 8);
		p = fnt;
		q = work + offbit + (jish * 2) + 256 * 16 * (0x7f-0x20);
		for (i = 0x01; i < 0x5f; i++) {
			p += 2;
			for (j = 16; j--;) {
				q -= 256;
				*(WORD *)q = ~(*(WORD *)(&p[j*256]));
			}
		}
	}
	p = work + offbit + 8 * 2 + 256 * 16 * (0x7f - 0x20);
	patchextank2bmp(p + 2, code29);
	patchextank2bmp(p + 4, code2a);
	patchextank2bmp(p + 6, code2b);
	patchextfnt2bmp(p + 8, code2c);

	p = work + offbit + 13 * 2 + 256 * 16 * (0x7f - 0x20);
	patchnecfnt2bmp(p, necfont);

	if ((fh = file_create_c(filename, FTYPE_BMP)) != (FILEH)-1) {
		file_lwrite(fh, work, size);
		file_close(fh);
	}

	free(work);
	gdk_pixmap_unref(font_pixmap);
	gdk_font_unref(fontset);
}
