/*	$Id: gtkmain.c,v 1.17 2003/07/22 14:55:58 nonaka Exp $	*/

/*
 * Copyright (c) 2002-2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "compiler.h"

#include <signal.h>

#include "np2.h"
#include "np2pc98.h"
#include "pc_core.h"
#include "iobridge.h"

#include "caption.h"
#include "diskdrv.h"
#include "dosio.h"
#include "ini.h"
#include "palettes.h"
#include "timing.h"
#include "trace.h"
#include "xnp2.h"
#include "xdraw.h"
#include "xkeyboard.h"
#include "xmenu.h"
#include "xmouse.h"

#include "icons/xnp2.xbm"

#ifndef CPU386
#define	APPNAME		"Xnp2"
#else
#define	APPNAME		"Xnp2+IA32"
#endif

#define	EVENT_MASK \
			(GDK_BUTTON1_MOTION_MASK	\
			 | GDK_BUTTON2_MOTION_MASK	\
			 | GDK_BUTTON3_MOTION_MASK	\
			 | GDK_POINTER_MOTION_MASK	\
			 | GDK_KEY_PRESS_MASK		\
			 | GDK_KEY_RELEASE_MASK		\
			 | GDK_BUTTON_PRESS_MASK	\
			 | GDK_BUTTON_RELEASE_MASK	\
			 | GDK_ENTER_NOTIFY_MASK	\
			 | GDK_LEAVE_NOTIFY_MASK	\
			 | GDK_EXPOSURE_MASK)

static void set_window_size(GtkWidget *);
static void set_icon_bitmap(GtkWidget *);
typedef	void sigfunc(int);
static sigfunc *setup_signal(int, sigfunc *);
static void sighandler(int);
static gint idle_process(gpointer *);

static int np2foreground = 0;

GtkWidget *window;
GtkWidget *main_vbox;
GtkWidget *menubar;
GtkWidget *drawarea;

WORD framecnt = 0;
WORD waitcnt = 0;
WORD framemax = 1;


/*
 - Signal: gint GtkWidget::expose_event (GtkWidget *WIDGET,
          GdkEventExpose *EVENT)
*/
static gint
expose(GtkWidget *w, GdkEventExpose *ev)
{

	if (ev->type == GDK_EXPOSE) {
		if (ev->count == 0) {
			xdraws_redraw(w);
		}
		return TRUE;
	}
	return FALSE;
}

/*
 - Signal: gint GtkWidget::key_press_event (GtkWidget *WIDGET,
          GdkEventKey *EVENT)
*/
static gint
key_press(GtkWidget *w, GdkEventKey *ev)
{

	if (ev->type == GDK_KEY_PRESS) {
		xkeyboard_down(w, ev->keyval);
		return TRUE;
	}
	return FALSE;
}

/*
 - Signal: gint GtkWidget::key_release_event (GtkWidget *WIDGET,
          GdkEventKey *EVENT)
*/
static gint
key_release(GtkWidget *w, GdkEventKey *ev)
{

	if (ev->type == GDK_KEY_RELEASE) {
		xkeyboard_up(w, ev->keyval);
		return TRUE;
	}
	return FALSE;
}

/*
 - Signal: gint GtkWidget::button_press_event (GtkWidget *WIDGET,
          GdkEventButton *EVENT)
*/
static gint
button_press(GtkWidget *w, GdkEventButton *ev)
{

	UNUSED(w);

	if (ev->type == GDK_BUTTON_PRESS) {
		switch (ev->button) {
		case 1:
			xmouse_button(MOUSE_LEFTDOWN);
			break;

		case 2:
			xmenu_toggle_item(MOUSE_MODE, !np2cfg.MOUSE_SW, TRUE);
			break;

		case 3:
			xmouse_button(MOUSE_RIGHTDOWN);
			break;
		}
		return TRUE;
	}
	return FALSE;
}

/*
 - Signal: gint GtkWidget::button_release_event (GtkWidget *WIDGET,
          GdkEventButton *EVENT)
*/
static gint
button_release(GtkWidget *w, GdkEventButton *ev)
{

	UNUSED(w);

	if (ev->type == GDK_BUTTON_RELEASE) {
		switch (ev->button) {
		case 1:
			xmouse_button(MOUSE_LEFTUP);
			break;

		case 2:
			break;

		case 3:
			xmouse_button(MOUSE_RIGHTUP);
			break;
		}
		return TRUE;
	}
	return FALSE;
}

/*
 - Signal: gint GtkWidget::enter_notify_event (GtkWidget *WIDGET,
          GdkEventCrossing *EVENT)
*/
static gint
enter_notify(GtkWidget *w, GdkEventCrossing *ev)
{

	if (ev->type == GDK_ENTER_NOTIFY) {
		np2foreground = 0;
		xdraws_drawall(w);
		return TRUE;
	}
	return FALSE;
}

/*
 - Signal: gint GtkWidget::leave_notify_event (GtkWidget *WIDGET,
          GdkEventCrossing *EVENT)
*/
static gint
leave_notify(GtkWidget *w, GdkEventCrossing *ev)
{

	UNUSED(w);

	if (ev->type == GDK_LEAVE_NOTIFY) {
		np2foreground = 1;
		return TRUE;
	}
	return FALSE;
}

/*
 - Signal: gint GtkWidget::delete_event (GtkWidget *WIDGET,
          GdkEvent *EVENT, GtkWidget *PARENT)
*/

/*
 * misc
 */
static void
set_window_size(GtkWidget *w)
{

	/* TCYs */
	gtk_window_set_policy(GTK_WINDOW(w), FALSE, FALSE, TRUE);
}

static void
set_icon_bitmap(GtkWidget *w)
{
	GdkPixmap *icon_pixmap;

	gdk_window_set_icon_name(w->window, APPNAME);
	icon_pixmap = gdk_bitmap_create_from_data(
	    w->window, np2_bits, np2_width, np2_height);
	gdk_window_set_icon(w->window, NULL, icon_pixmap, NULL);
}

#if 0
static GtkWidget* create_pixmap(GtkWidget *, gchar **);
static GtkWidget *
create_pixmap(GtkWidget *widget, gchar **data)
{
	GtkWidget *pixmap;
	GdkColormap *colormap;
	GdkPixmap *gdkpixmap;
	GdkBitmap *mask;

	colormap = gtk_widget_get_colormap(widget);
	gdkpixmap = gdk_pixmap_colormap_create_from_xpm_d(
	    NULL, colormap, &mask, NULL, data);
	if (gdkpixmap == NULL) {
		g_warning("Couldn't create pixmap.");
		return NULL;
	}

	pixmap = gtk_pixmap_new(gdkpixmap, mask);
	gdk_pixmap_unref(gdkpixmap);
	gdk_bitmap_unref(mask);

	return pixmap;
}
#endif

void
set_window_title(char *titlestr)
{

	gtk_window_set_title(GTK_WINDOW(window), titlestr);
}

BYTE
changescreen(GtkWidget *w, BYTE mode)
{

        if (mode != 0 && xdraws_setscreenmode(mode) == 0) {
                return(SUCCESS);
        }
        audio_stop();
        mouse_running(w, MOUSE_STOP);
        xdraws_term(w);
        xdraws_windowstats(w);
        if (xdraws_setup(w)) {
		return FAILURE;
        }
        xdraws_redraw(w);
        mouse_running(w, MOUSE_CONT);
        audio_play();
        return SUCCESS;
}

/*
 * signal handler
 */
static sigfunc *
setup_signal(int signo, sigfunc *func)
{
	struct sigaction act, oact;

	act.sa_handler = func;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	if (sigaction(signo, &act, &oact) < 0)
		return SIG_ERR;
	return oact.sa_handler;
}

static void
sighandler(int signo)
{

	UNUSED(signo);

	gtk_main_quit();
}

/*
 * NP2
 */
void
np2os_exit(void)
{

	gtk_main_quit();
}


/*
 * main
 */
int
main(int argc, char *argv[])
{
	int rv = 1;

	if (set_modulepath(modulefile, sizeof(modulefile)))
		return 1;

	dosio_init();
	file_setcd(modulefile);
	initload();

	gtk_set_locale();
	gtk_rc_add_default_file(".xnp2rc");
	gtk_init(&argc, &argv);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

	main_vbox = gtk_vbox_new(FALSE, 1);
	gtk_container_border_width(GTK_CONTAINER(main_vbox), 1);
	gtk_container_add(GTK_CONTAINER(window), main_vbox);
	gtk_widget_show(main_vbox);

	/* j[o[	*/
	menubar = create_menu(window);
	gtk_box_pack_start(GTK_BOX(main_vbox), menubar, FALSE, TRUE, 0);
	gtk_widget_show(menubar);

	/* ʗ̈ */
	drawarea = gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(drawarea),
	    SCREEN_WIDTH, SCREEN_HEIGHT);
	gtk_box_pack_start(GTK_BOX(main_vbox), drawarea, FALSE, TRUE, 0);
	gtk_widget_show(drawarea);

	gtk_widget_realize(window);

	/* ^CgACR̃ZbgAbv */
	gtk_window_set_title(GTK_WINDOW(window), np2cfg.titles);
	set_window_size(window);
	set_icon_bitmap(window);

	/* setup signal */
	gtk_signal_connect(GTK_OBJECT(window), "destroy", 
	    GTK_SIGNAL_FUNC(gtk_main_quit), "WM destroy");
	gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
	    GTK_SIGNAL_FUNC(key_press), NULL);
	gtk_signal_connect(GTK_OBJECT(window), "key_release_event",
	    GTK_SIGNAL_FUNC(key_release), NULL);
	gtk_signal_connect(GTK_OBJECT(window), "button_press_event",
	    GTK_SIGNAL_FUNC(button_press), NULL);
	gtk_signal_connect(GTK_OBJECT(window), "button_release_event",
	    GTK_SIGNAL_FUNC(button_release), NULL);

	/* setup drawarea signal */
	gtk_signal_connect(GTK_OBJECT(drawarea), "expose_event",
	    GTK_SIGNAL_FUNC(expose), NULL);
	gtk_signal_connect(GTK_OBJECT(drawarea), "enter_notify_event",
	    GTK_SIGNAL_FUNC(enter_notify), NULL);
	gtk_signal_connect(GTK_OBJECT(drawarea), "leave_notify_event",
	    GTK_SIGNAL_FUNC(leave_notify), NULL);

	gtk_widget_add_events(window, EVENT_MASK);

	TRACE_START();

	caption_disp(1);

	xdraws_setscreenmode(SCMD_SETWINDOWED);
	xdraws_windowstats(drawarea);

	np2pal_initpalettes();
	draws_init();
	if (xdraws_setup(drawarea))
		goto resource_cleanup;

	/* window \ */
	gtk_widget_show(window);

	INIT_RESET;

	np2os_init();
	np2_init();

	np2os_reset();
	np2_reset();

	xdraws_redraw(drawarea);

	np2_reset();

	setup_signal(SIGINT, sighandler);
	setup_signal(SIGTERM, sighandler);

	/* C[v */
	install_idle_process();
	gtk_main();
	uninstall_idle_process();
	rv = 0;

	if (np2_getmemsw())
		INIT_RENEWAL;

	np2_term();
	np2os_term();

resource_cleanup:
	xdraws_term(drawarea);
	initsave();

	TRACE_END();
	dosio_term();

	return rv;
}

static gint
idle_process(gpointer *p)
{
	GtkWidget *w = (GtkWidget *)p;

	if (np2cfg.NOWAIT) {
		mouse_callback(w);
		np2main_exec(!framecnt);
		caption_fddstat();
		if (np2cfg.DRAW_SKIP) {
			framecnt++;
			if (framecnt >= np2cfg.DRAW_SKIP) {
				processwait(0);
			}
		} else {
			framecnt = 1;
			if (timing_getcount()) {
				processwait(0);
			}
		}
	} else if (np2cfg.DRAW_SKIP) {
		if (framecnt < np2cfg.DRAW_SKIP) {
			mouse_callback(w);
			np2main_exec(!framecnt);
			caption_fddstat();
			framecnt++;
		} else {
			processwait(np2cfg.DRAW_SKIP);
		}
	} else {
		if (waitcnt == 0) {
			mouse_callback(w);
			np2main_exec(!framecnt);
			caption_fddstat();
			framecnt++;
			if (timing_getcount() < framecnt) {
				waitcnt = framecnt;
				if (framemax > 1) {
					framemax--;
				}
			} else if (framecnt >= framemax) {
				waitcnt = framecnt;
				if (framemax < 12) {
					framemax++;
				}
			}
		} else {
			processwait(waitcnt);
			waitcnt = framecnt;
		}
	}

	return TRUE;
}

static int idle_id;

void
install_idle_process(void)
{

	idle_id = gtk_idle_add((GtkFunction)idle_process, drawarea);
	audio_play();
}

void
uninstall_idle_process(void)
{

	audio_stop();
	gtk_idle_remove(idle_id);
}
