/*	$Id: gtkmouse.c,v 1.8 2003/06/05 18:11:23 nonaka Exp $	*/

/*
 * Copyright (c) 2002-2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "compiler.h"

#include "np2.h"
#include "xnp2.h"
#include "xmouse.h"

BYTE mouse_move_ratio = MOUSE_RATIO_100;
BYTE mouse_move_mul;
BYTE mouse_move_div;

static GdkPixmap *cursor_pixmap;
static GdkCursor *cursor;

static int mouserunning = 0;	// showing
static int lastmouse = 0;	// working
static short mousex = 0;
static short mousey = 0;
static BYTE mouseb = 0xa0;

static void getmaincenter(GtkWidget *, int *, int *);
static void mouseonoff(GtkWidget *, int);

int
xmouse_init(GtkWidget *w)
{
	static gchar hide_cursor[16*16/8] = {
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	};

	cursor_pixmap = gdk_pixmap_create_from_data(w->window, hide_cursor,
	    16, 16, 1, &w->style->black, &w->style->black);
	cursor = gdk_cursor_new_from_pixmap(cursor_pixmap, cursor_pixmap,
	    &w->style->black, &w->style->black, 0, 0);

	return 0;
}

void
xmouse_term(GtkWidget *w)
{

	UNUSED(w);

	if (cursor_pixmap)
		gdk_pixmap_unref(cursor_pixmap);
}

static void
getmaincenter(GtkWidget *w, int *cx, int *cy)
{

	*cx = w->allocation.x + w->allocation.width / 2;
	*cy = w->allocation.y + w->allocation.height / 2;
}

static void
mouseonoff(GtkWidget *w, int flag)
{

	if ((lastmouse ^ flag) & 1) {
		lastmouse = flag & 1;
		if (lastmouse) {
			int curx, cury;

			gdk_pointer_grab(w->window, TRUE, 0, w->window,
			    cursor, 0);
			getmaincenter(w, &curx, &cury);
			gdk_window_set_pointer(w->window, curx, cury);
		} else {
			gdk_pointer_ungrab(0);
		}
	}
}

void
xmouse_show(GtkWidget *w, int flag)
{

	if (flag == MOUSE_TOGGLE) {
		mouseonoff(w, mouserunning ^ 1);
		mouserunning ^= 1;
	} else {
		if (mouserunning ^ flag) {
			mouseonoff(w, flag);
			mouserunning = flag & 1;
		}
	}
}

void
mouse_running(GtkWidget *w, BYTE flg)
{
	BYTE mf = mouserunning;

	switch (flg & 0xc0) {
	case 0x00:
		mf &= ~(1 << (flg & 7));
		break;

	case 0x40:
		mf ^= (1 << (flg & 7));
		break;

	default:
		mf |= (1 << (flg & 7));
		break;
	}

	if ((mf ^ mouserunning) & MOUSE_MASK) {
		mouserunning = (mf & MOUSE_MASK);
		mouseonoff(w, (mouserunning == 1) ? 1 : 0);
	}
}

void
mouse_callback(GtkWidget *w)
{
	int wx, wy;
	int cx, cy;

	if (lastmouse & 1) {
		gdk_window_get_pointer(w->window, &wx, &wy, NULL);
		getmaincenter(w, &cx, &cy);
		mousex += (short)((wx - cx) / 2);
		mousey += (short)((wy - cy) / 2);
		gdk_window_set_pointer(w->window, cx, cy);
	}
}

int
xmouse_button(int button)
{

	if ((lastmouse & 1) == 0)
		return 0;

	switch (button) {
	case MOUSE_LEFTDOWN:
		mouseb &= 0x7f;
		break;

	case MOUSE_LEFTUP:
		mouseb |= 0x80;
		break;

	case MOUSE_RIGHTDOWN:
		mouseb &= 0xdf;
		break;

	case MOUSE_RIGHTUP:
		mouseb |= 0x20;
		break;
	}
	return 1;
}

BYTE
np2os_getmousestat(short *x, short *y, int clear)
{

	if (mouse_move_ratio == MOUSE_RATIO_100) {
		*x = mousex;
		*y = mousey;
	} else if (mouse_move_div == 1) {
		*x = mousex * mouse_move_mul;
		*y = mousey * mouse_move_mul;
	} else {
		*x = (mousex * mouse_move_mul) / mouse_move_div;
		*y = (mousey * mouse_move_mul) / mouse_move_div;
	}
	if (clear) {
		mousex = 0;
		mousey = 0;
	}
	return mouseb;
}
