#include	"compiler.h"

#include	"i286.h"
#include	"memory.h"
#include	"iobridge.h"
#include	"dmac.h"
#include	"fdc.h"
#include	"nevent.h"
#include	"pic.h"
#include	"trace.h"

	DMA_TABLE	dma;

static DMAC_OUT dummy_out(BYTE data) { }
static DMAC_IN dummy_in(void) { return(0xff); }
static DMAC_EXT dummy_dmaproc(BYTE func) { return(0); }

void dmac_init(void) {

	int		i;

	ZeroMemory(&dma, sizeof(dma));
	dma.mask = 0xf;
	for (i=0; i<4; i++) {
		dma.dmach[i].outproc = dummy_out;
		dma.dmach[i].inproc = dummy_in;
		dma.dmach[i].extproc = dummy_dmaproc;
	}
	dma.dmach[0].extproc = cs4231dmafunc;
	dma.dmach[DMA_2HD].inproc = fdc_DataRegRead;
	dma.dmach[DMA_2HD].outproc = fdc_DataRegWrite;
	dma.dmach[DMA_2HD].extproc = fdc_dmafunc;
	dma.dmach[DMA_2DD].inproc = fdc_DataRegRead;
	dma.dmach[DMA_2DD].outproc = fdc_DataRegWrite;
	dma.dmach[DMA_2DD].extproc = fdc_dmafunc;
}



void dmac_check(void) {

	int		i;
	BOOL	workchg = FALSE;
	BYTE	bit;

	for (i=0, bit=1; i<4; i++, bit<<=1) {
		if ((!(dma.mask & bit)) && (dma.dmach[i].ready)) {
			if (!(dma.work & bit)) {
				dma.work |= bit;
				if (dma.dmach[i].extproc(DMAEXT_START)) {
					dma.stat &= ~bit;						// ver0.27
					dma.working |= bit;
					workchg = TRUE;
				}
			}
		}
		else {
			if (dma.work & bit) {
				dma.work &= ~bit;
				dma.working &= ~bit;
				dma.dmach[i].extproc(DMAEXT_BREAK);
				workchg = TRUE;
			}
		}
	}
	if (workchg) {
		nevent_forceexit();
	}
}


I286_OUT dmac_w(WORD port, BYTE value) {

	if (!(port & 1)) {
		return;
	}
//	TRACEOUT(port, value);
	if (port < 0x10) {
		if (port & 2) {
			XWORD_BYTE(dma.dmach[port >> 2].leng, dma.lh) = value;
			XWORD_BYTE(dma.dmach[port >> 2].lengorg, dma.lh) = value;
			dma.stat &= ~(1 << (port >> 2));					// ver0.27
		}
		else {
			XDWORD_BYTE(dma.dmach[port >> 2].adrs, X_LOW, dma.lh) = value;
			XWORD_BYTE(dma.dmach[port >> 2].adrsorg, dma.lh) = value;
		}
		dma.lh ^= 1;
	}
	else {
		switch(port) {
			case 0x11:
				break;
			case 0x13:
				dma.dmach[value & 3].sreq = value;
				break;
			case 0x15:
				if (value & 4) {
					dma.mask |= (1 << (value & 3));
				}
				else {
					dma.mask &= ~(1 << (value & 3));
				}
				dmac_check();
				break;
			case 0x17:
				dma.dmach[value & 3].mode = value;
				break;
			case 0x19:
				dma.lh = 0;
				break;
			case 0x1b:
				dma.mask = 0x0f;
				break;
			case 0x1d:
				break;
			case 0x1f:
				dma.mask = value;
				dmac_check();
				break;
		}
	}
}

I286_INB dmac_r(WORD port) {

	BYTE	ret;

	if (port & 0x01) {
		TRACEIN(port);
		if (port < 0x10) {
			if (port & 2) {
				ret = XDWORD_BYTE(dma.dmach[port >> 2].adrs, X_LOW, dma.lh);
			}
			else {
				ret = XWORD_BYTE(dma.dmach[port >> 2].leng, dma.lh);
			}
			dma.lh ^= 1;
			return(ret);
		}
		else {
			switch(port) {									// ver0.27
				case 0x11:
					return(dma.stat);						// ToDo!!
				case 0x1b:
					break;
			}
		}
	}
	return(0xff);
}


I286_OUT dmabank_w(WORD port, BYTE value) {

	if (port & 1) {
		if (port < 0x28) {
			XDWORD_WORD(dma.dmach[((port+2)>>1)&3].adrs, X_HIGH) = value;
		}
		else if (port == 0x29) {
		}
	}
}

I286_INB dmabank_r(WORD port) {

	return(0xff);
}



// --------------------------------------------------------------------------

#ifdef NP2_DMAC_ASM

LABEL void dma_proc(void) {			// called by i286 : regist esi, ebp

		__asm {
				cmp		dma.working, 0
				je		noworkdma
				pushad										// for C
				mov		edi, offset (dma.dmach)
				mov		cl, 4
				mov		ebx, 1
searchworkdma:	test	dma.working, ebx
				jne		hitworkdma
				add		ebx, ebx
				add		edi, type DMACH_TABLE
				dec		cl
				jne		searchworkdma
				popad										// for C
noworkdma:		ret

hitworkdma:		sub		word ptr ([edi]DMACH_TABLE.leng.w), 1
				jnc		dmaprocs
				or		dma.stat, bl							// ver0.27
				not		ebx
				and		dma.working, ebx
				mov		ecx, DMAEXT_END
				call	[edi]DMACH_TABLE.extproc

dmaprocs:		mov		al, [edi]DMACH_TABLE.mode
				and		al, 0ch
				je		dma_verify
				cmp		al, 04h
				je		dma_port2mem

				mov		ecx, dword ptr ([edi]DMACH_TABLE.adrs.d)
				call	i286_memoryread
				mov		cl, al
				call	[edi]DMACH_TABLE.outproc
				jmp		dma_counters

dma_verify:		call	[edi]DMACH_TABLE.inproc
				jmp		dma_counters

dma_port2mem:	call	[edi]DMACH_TABLE.inproc
				mov		ecx, dword ptr ([edi]DMACH_TABLE.adrs.d)
				mov		dl, al
				call	i286_memorywrite

dma_counters:	test	[edi]DMACH_TABLE.mode, 20h
				jne		short dmaadrsdec
				inc		word ptr ([edi]DMACH_TABLE.adrs.x[0].w)
				popad
				ret											// for C
dmaadrsdec:		dec		word ptr ([edi]DMACH_TABLE.adrs.x[0].w)
				popad										// for C
				ret
	}
}

#else

void dma_proc(void) {

	DMACH_TABLE	*dch;
	int			bit;

	if (dma.working) {
		dch = dma.dmach;
		bit = 1;
		do {
			if (dma.working & bit) {
				// DMA working !
				if (!dch->leng.w) {
					dma.stat |= bit;
					dma.working &= ~bit;
					dch->extproc(DMAEXT_END);
				}
				dch->leng.w--;

				switch((dch->mode) & 0x0c) {
					case 0x00:		// verifty
						dch->inproc();
						break;

					case 0x04:		// port->mem
						i286_memorywrite(dch->adrs.d, dch->inproc());
						break;

					default:
						dch->outproc(i286_memoryread(dch->adrs.d));
						break;
				}
				// ver0.28
				XDWORD_DWORD(dch->adrs) += (((dch->mode) & 0x20)?-1:1);
			}
			dch++;
			bit <<= 1;
		} while(bit & 0x0f);
	}
}

#endif

