#include	"compiler.h"

#include	"np2.h"
#include	"pc_core.h"
#include	"palettes.h"
#include	"iobridge.h"
#include	"gdc.h"

		PALETTE_TABLE	np2_pal32[256];
		WORD			np2_pal16[256];
		PAL_EVENT		palevent;
static	PALETTE_TABLE	degpal1[8];
static	PALETTE_TABLE	degpal2[8];
static	BYTE			anapal1[16];
static	BYTE			anapal2[16];

static	PALETTE_TABLE	lcdpal[15];
static	BYTE			lcdtbl[0x1000];
static	PALETTE_TABLE	pal16bit;
static	BYTE			np2_r16b;
static	BYTE			np2_l16r;
static	BYTE			np2_l16g;
static	int				np2_palbits = 8;


static	DWORD	degtbl[8] = {	DWORD_RGB(0,0,0), DWORD_RGB(0,0,1),
								DWORD_RGB(0,1,0), DWORD_RGB(0,1,1),
								DWORD_RGB(1,0,0), DWORD_RGB(1,0,1),
								DWORD_RGB(1,1,0), DWORD_RGB(1,1,1)};

static	BYTE	lcdpal_a[27] = {0, 1, 2, 3, 5, 2, 4, 4, 6,
								7, 9, 2,11,13, 2, 4, 4, 6,
								8, 8,10, 8, 8,10,12,12,14};



void np2pal_make16mask(DWORD bmask, DWORD rmask, DWORD gmask) {

	WORD	bit;
	BYTE	cnt;

	pal16bit.d = 0;
	for (bit=1; (bit) && (!(bmask & bit)); bit<<=1) { }
	for (np2_r16b=8; (np2_r16b) && (bmask & bit); np2_r16b--, bit<<=1) {
		pal16bit.p.b >>= 1;
		pal16bit.p.b |= 0x80;
	}
	for (np2_l16r=0, bit=1; bit && (!(rmask & bit)); np2_l16r++, bit<<=1) { }
	for (cnt=0x80; (cnt) && (rmask & bit); cnt>>=1, bit<<=1) {
		pal16bit.p.r |= cnt;
	}
	for (; cnt; cnt>>=1) {
		np2_l16r--;
	}
	for (np2_l16g=0, bit=1; bit && (!(gmask & bit)); np2_l16g++, bit<<=1) { }
	for (cnt=0x80; (cnt) && (gmask & bit); cnt>>=1, bit<<=1) {
		pal16bit.p.g |= cnt;
	}
	for (; cnt; cnt>>=1) {
		np2_l16g--;
	}
}

WORD np2pal_get16pal(DWORD pal) {

	PALETTE_TABLE	ret;

	ret.d = pal & pal16bit.d;
	return((((WORD)ret.p.g) << np2_l16g) |
			(((WORD)ret.p.r) << np2_l16r) | (ret.p.b >> np2_r16b));
}


void np2pal_makegrad(PALETTE_TABLE *pal, int pals, DWORD bg, DWORD fg) {

	int		i;

	if (pals >= 2) {
		pals--;
		for (i=0; i<=pals; i++) {
			pal[i].p.b = (BYTE)
				((((fg >> 0) & 0x0000ff) * i + 
				((bg >> 0) & 0x0000ff) * (pals-i)) / pals);
			pal[i].p.g = (BYTE)
				((((fg >> 8) & 0x0000ff) * i + 
				((bg >> 8) & 0x0000ff) * (pals-i)) / pals);
			pal[i].p.r = (BYTE)
				((((fg >> 16) & 0x0000ff) * i + 
				((bg >> 16) & 0x0000ff) * (pals-i)) / pals);
			pal[i].p.f = 0;
		}
	}
}

// ---------------------------------------------------------------------------

void pal_initlcdtable(void) {

	DWORD	i;
	int		j;

	for (i=0; i<0x1000; i++) {
		j = 0;
		if ((i & 0x00f) >= 0x004) {					// b
			j++;
			if ((i & 0x00f) >= 0x00b) {
				j++;
			}
		}
		if ((i & 0x0f0) >= 0x040) {					// r
			j += 3;
			if ((i & 0x0f0) >= 0x0b0) {
				j += 3;
			}
		}
		if ((i & 0xf00) >= 0x400) {					// g
			j += 9;
			if ((i & 0xf00) >= 0xb00) {
				j += 9;
			}
		}
		lcdtbl[i] = lcdpal_a[j];
	}
}

void pal_makelcdpal(void) {

	if (!(np2cfg.LCD_MODE & 2)) {
		np2pal_makegrad(lcdpal, 15, np2cfg.BG_COLOR, np2cfg.FG_COLOR);
	}
	else {
		np2pal_makegrad(lcdpal, 15, np2cfg.FG_COLOR, np2cfg.BG_COLOR);
	}
}

void pal_makeskiptable(void) {

	int		i;

	for (i=0; i<8; i++) {
		degpal1[i].d = degtbl[i] * 255;
		degpal2[i].d = degtbl[i] * np2cfg.skiplight;
	}
	for (i=0; i<16; i++) {
		anapal1[i] = (BYTE)(i * 0x11);
		anapal2[i] = (BYTE)((np2cfg.skiplight * anapal1[i]) / 255);
	}
}


// ---------------------------------------------------------------------------

void pal_makeanalog(PALETTE_TABLE *pal, WORD bit) {

	WORD	i;

	for (i=0; i<NP2PALS_GRPH; i++, pal++) {
		if (bit & (1 << i)) {
			np2_pal32[i+NP2PAL_GRPH].p.b = anapal1[pal->p.b & 15];
			np2_pal32[i+NP2PAL_GRPH].p.g = anapal1[pal->p.g & 15];
			np2_pal32[i+NP2PAL_GRPH].p.r = anapal1[pal->p.r & 15];
			if (np2cfg.skipline) {
				np2_pal32[i+NP2PAL_SKIP].p.b = anapal2[pal->p.b & 15];
				np2_pal32[i+NP2PAL_SKIP].p.g = anapal2[pal->p.g & 15];
				np2_pal32[i+NP2PAL_SKIP].p.r = anapal2[pal->p.r & 15];
			}
		}
	}
	if (np2_palbits == 16) {
		for (i=0; i<NP2PALS_GRPH; i++) {
			if (bit & (1 << i)) {
				np2_pal16[i+NP2PAL_GRPH] =
								np2pal_get16pal(np2_pal32[i+NP2PAL_GRPH].d);
				np2_pal16[i+NP2PAL_SKIP] =
								np2pal_get16pal(np2_pal32[i+NP2PAL_SKIP].d);
			}
		}
	}
}

static void pal_makedegital(void) {

	WORD	i;

	for (i=0; i<4; i++) {
		np2_pal32[i+NP2PAL_GRPH+ 0].d =
		np2_pal32[i+NP2PAL_GRPH+ 8].d =
									degpal1[(gdc.degpal[i] >> 4) & 7].d;
		np2_pal32[i+NP2PAL_GRPH+ 4].d =
		np2_pal32[i+NP2PAL_GRPH+12].d =
									degpal1[gdc.degpal[i] & 7].d;
		if (np2cfg.skipline) {
			np2_pal32[i+NP2PAL_SKIP+ 0].d =
			np2_pal32[i+NP2PAL_SKIP+ 8].d =
									degpal2[(gdc.degpal[i] >> 4) & 7].d;
			np2_pal32[i+NP2PAL_SKIP+ 4].d =
			np2_pal32[i+NP2PAL_SKIP+12].d =
									degpal2[gdc.degpal[i] & 7].d;
		}
	}
	if (np2_palbits == 16) {
		for (i=0; i<4; i++) {
			np2_pal16[i+NP2PAL_GRPH+ 0] =
			np2_pal16[i+NP2PAL_GRPH+ 8] =
								np2pal_get16pal(np2_pal32[i+NP2PAL_GRPH+0].d);
			np2_pal16[i+NP2PAL_GRPH+ 4] =
			np2_pal16[i+NP2PAL_GRPH+12] =
								np2pal_get16pal(np2_pal32[i+NP2PAL_GRPH+4].d);
		}
		if (np2cfg.skipline) {
			for (i=0; i<4; i++) {
				np2_pal16[i+NP2PAL_SKIP+ 0] =
				np2_pal16[i+NP2PAL_SKIP+ 8] =
								np2pal_get16pal(np2_pal32[i+NP2PAL_SKIP+0].d);
				np2_pal16[i+NP2PAL_SKIP+ 4] =
				np2_pal16[i+NP2PAL_SKIP+12] =
								np2pal_get16pal(np2_pal32[i+NP2PAL_SKIP+4].d);
			}
		}
	}
}

void pal_makeanalog_lcd(PALETTE_TABLE *pal, WORD bit) {

	WORD	i;
	DWORD	j;

	for (i=0; i<NP2PALS_GRPH; i++, pal++) {
		if (bit & (1 << i)) {
			j = (pal->p.b & 15);
			j |= (pal->p.r & 15) << 4;
			j |= (pal->p.g & 15) << 8;
			np2_pal32[i+NP2PAL_SKIP].d =
			np2_pal32[i+NP2PAL_GRPH].d = lcdpal[lcdtbl[j]].d;
		}
	}
	if (np2_palbits == 16) {
		for (i=0; i<NP2PALS_GRPH; i++) {
			if (bit & (1 << i)) {
				np2_pal16[i+NP2PAL_GRPH] =
				np2_pal16[i+NP2PAL_SKIP] =
								np2pal_get16pal(np2_pal32[i+NP2PAL_GRPH].d);
			}
		}
	}
}

static void pal_makedegital_lcd(void) {

	WORD	i;

	for (i=0; i<4; i++) {
		np2_pal32[i+NP2PAL_GRPH+ 0].d =
		np2_pal32[i+NP2PAL_GRPH+ 8].d =
									lcdpal[(gdc.degpal[i] >> 3) & 14].d;
		np2_pal32[i+NP2PAL_GRPH+ 4].d =
		np2_pal32[i+NP2PAL_GRPH+12].d =
									lcdpal[(gdc.degpal[i] << 1) & 14].d;
		if (np2cfg.skipline) {
			np2_pal32[i+NP2PAL_SKIP+ 0].d =
			np2_pal32[i+NP2PAL_SKIP+ 8].d = np2_pal32[i+NP2PAL_GRPH+ 0].d;
			np2_pal32[i+NP2PAL_SKIP+ 4].d =
			np2_pal32[i+NP2PAL_SKIP+12].d = np2_pal32[i+NP2PAL_GRPH+ 4].d;
		}
	}
	if (np2_palbits == 16) {
		for (i=0; i<4; i++) {
			np2_pal16[i+NP2PAL_GRPH+ 0] =
			np2_pal16[i+NP2PAL_GRPH+ 8] =
								np2pal_get16pal(np2_pal32[i+NP2PAL_GRPH+0].d);
			np2_pal16[i+NP2PAL_GRPH+ 4] =
			np2_pal16[i+NP2PAL_GRPH+12] =
								np2pal_get16pal(np2_pal32[i+NP2PAL_GRPH+4].d);
		}
		if (np2cfg.skipline) {
			for (i=0; i<4; i++) {
				np2_pal16[i+NP2PAL_SKIP+ 0] =
				np2_pal16[i+NP2PAL_SKIP+ 8] = np2_pal16[i+NP2PAL_GRPH+ 0];
				np2_pal16[i+NP2PAL_SKIP+ 4] =
				np2_pal16[i+NP2PAL_SKIP+12] = np2_pal16[i+NP2PAL_GRPH+ 4];
			}
		}
	}
}

static void pal_maketext(void) {

	int		i, j, k;
	WORD	pal16;

	k = NP2PAL_TEXT2;
	for (i=0; i<8; i++) {
		np2_pal32[i+1+NP2PAL_TEXT].d = degpal1[i].d;
		np2_pal32[i+1+NP2PAL_TEXT3].d = degpal1[i].d;
		for (j=0; j<NP2PALS_GRPH; j++, k++) {
			np2_pal32[k].d = degpal1[i].d;
		}
	}
	np2_pal32[NP2PAL_TEXT3] = np2_pal32[NP2PAL_TEXT3 + 1];
	if (np2_palbits == 16) {
		k = NP2PAL_TEXT2;
		for (i=0; i<8; i++) {
			pal16 = np2pal_get16pal(degpal1[i].d);
			np2_pal16[i+1+NP2PAL_TEXT] = pal16;
			np2_pal16[i+1+NP2PAL_TEXT3] = pal16;
			for (j=0; j<NP2PALS_GRPH; j++, k++) {
				np2_pal16[k] = pal16;
			}
		}
		np2_pal16[NP2PAL_TEXT3] = np2_pal16[NP2PAL_TEXT3 + 1];
	}
}

static void pal_maketext_lcd(void) {

	int		i, j, k;
	WORD	pal16;

	k = NP2PAL_TEXT2;
	for (i=0; i<8; i++) {
		np2_pal32[i+1+NP2PAL_TEXT].d = lcdpal[i*2].d;
		np2_pal32[i+1+NP2PAL_TEXT3].d = lcdpal[i*2].d;
		for (j=0; j<NP2PALS_GRPH; j++, k++) {
			np2_pal32[k].d = lcdpal[i*2].d;
		}
	}
	np2_pal32[NP2PAL_TEXT3] = np2_pal32[NP2PAL_TEXT3 + 1];
	if (np2_palbits == 16) {
		k = NP2PAL_TEXT2;
		for (i=0; i<8; i++) {
			pal16 = np2pal_get16pal(lcdpal[i*2].d);
			np2_pal16[i+1+NP2PAL_TEXT] = pal16;
			np2_pal16[i+1+NP2PAL_TEXT3] = pal16;
			for (j=0; j<NP2PALS_GRPH; j++, k++) {
				np2_pal16[k] = pal16;
			}
		}
		np2_pal16[NP2PAL_TEXT3] = np2_pal16[NP2PAL_TEXT3 + 1];
	}
}

void pal_change(BYTE textpalset) {

	if (textpalset) {
		if (!(np2cfg.LCD_MODE & 1)) {
			pal_maketext();
		}
		else {
			pal_maketext_lcd();
		}
		np2_pal32[NP2PAL_TEXT].d = np2_pal32[NP2PAL_TEXT2].d;
		np2_pal16[NP2PAL_TEXT] = np2_pal16[NP2PAL_TEXT2];
	}
	if (!(np2cfg.LCD_MODE & 1)) {
		if (gdc.analog) {
			pal_makeanalog(gdc.anapal, 0xffff);
		}
		else {
			pal_makedegital();
		}
	}
	else {
		if (gdc.analog) {
			pal_makeanalog_lcd(gdc.anapal, 0xffff);
		}
		else {
			pal_makedegital_lcd();
		}
	}
	if (np2cfg.skipline) {
		np2_pal32[NP2PAL_TEXT].d = np2_pal32[NP2PAL_SKIP].d;
		np2_pal16[NP2PAL_TEXT] = np2_pal16[NP2PAL_SKIP];
	}
	np2os_changepalettes();
}


void pal_eventclear(void) {

extern	WORD	framecnt;

	palevent.anabit = 0;
	palevent.events = 0;
	if ((!np2scrnoutflg) || (!np2cfg.RASTER) || (np2_palbits == 8)) {
		palevent.events--;					// 0xffffffff ˤ...
	}
	else {
		CopyMemory(palevent.pal, gdc.anapal, sizeof(gdc.anapal));
	}
}

// ------------------------------------------------


void np2pal_initpalettes(void) {

	pal_initlcdtable();
	pal_makelcdpal();
	pal_makeskiptable();
}

void np2pal_setcolorbit(int bit) {

	np2_palbits = bit;
}

