#include	"compiler.h"

#include	"dosio.h"
#include	"iobridge.h"
#include	"sndboard.h"
#include	"pcm86.h"
#include	"np2sound.h"

	PCM86_TABLE		pcm86;

//									44  33  22  16  11  08  05  04
const	BYTE	p86rate_tbl[] = {   32, 24, 16, 12,  8,  6,  4,  3};
		long	pcm86rate8[] = {352800, 264600, 176400, 132300,
								 88200,  66150,  44010,  33075};

void pcm86_reset(void) {									// ver0.27a

	ZeroMemory(&pcm86, sizeof(pcm86));
	pcm86.step = (BYTE)(32 * np2sound.pcm86rate);
	if (np2sound.rate) {
		pcm86.div = (44100 << (PCM86_DIVBIT - 3)) / np2sound.rate;
		pcm86.div2 = (np2sound.rate << (PCM86_DIVBIT + 3)) / 44100;
	}
}

I286_INB pcm86_r(WORD port) {

	BYTE	ret = 0;

	switch(port) {
		case 0xa460:
			ret = 0x40 | (pcm86.extfunc & 1);
			break;

		case 0xa466:
			sound_make();
			ret = pcm86.lrclock & 1;
			if (pcm86.bufsize >= 0x8000U) {		// Хåեե
				ret |= 0x80;
			}
			else if (!pcm86.bufsize) {			// Хåե
				ret |= 0x40;
			}
			break;
		case 0xa468:							// FIFO
			ret = pcm86.fifo & (~0x10);
			if ((pcm86.write) && (pcm86.fifo & 0x20)) {
//				pcm86.write = 0;
				sound_make();
				if (pcm86.bufsize < pcm86.fifosize) {
					pcm86.write = 0;
					ret |= 0x10;
				}
			}
			break;
		case 0xa46a:							// D/A
			ret = pcm86.dactrl;
			break;
#if 0
		case 0xa46c:							// FIFOǡ
			break;
#endif
	}
	return(ret);
}


I286_OUT pcm86_w(WORD port, BYTE val) {

	switch(port) {
		case 0xa460:
			pcm86.extfunc = val;
			if (sndproc.extendchannel) {
				sndproc.extendchannel(val & 1);
			}
			break;
#if 0
		case 0xa466:								// ܥ塼
			break;
#endif
		case 0xa468:								// FIFO
			if ((!(pcm86.fifo & 8)) && (val & 8)) {
				pcm86.readpos = 0;				// Хåեꥻå
				pcm86.bufsize = 0;
				pcm86.count = 0;
				pcm86.lrclock = 0;
			}
			pcm86.step = (BYTE)(p86rate_tbl[val & 7] * np2sound.pcm86rate);
			if (np2sound.rate) {
				long rate = pcm86rate8[val & 7];
				pcm86.div = (rate << (PCM86_DIVBIT - 3)) / np2sound.rate;
				pcm86.div2 = (np2sound.rate << (PCM86_DIVBIT + 3)) / rate;
			}
			pcm86.fifo = val & (~0x10);
			pcm86.write = 1;
			break;
		case 0xa46a:								// D/A
			if (pcm86.fifo & 0x20) {
				pcm86.fifosize = (WORD)(val) << 7;
			}
			else {
				pcm86.dactrl = val;
			}
			break;
		case 0xa46c:							// FIFOǡ
			if (pcm86.bufsize < 0x8000U) {
				pcm86.buffer[(pcm86.readpos + pcm86.bufsize) & 0x7fff] = val;
				pcm86.bufsize++;
				pcm86.write = 1;
				if (pcm86.dactrl & 0x40) {
					pcm86.lrclock ^= 1;
				}
				else {
					pcm86.lrclock ^= 2;
					if (!(pcm86.lrclock & 2)) {
						pcm86.lrclock ^= 1;
					}
				}
			}
			break;
	}
}




#ifdef SOUND_86PCM_ASM

LABEL void __fastcall pcm86_update(void *buf, int buflength) {

		__asm {
				cmp		edx, 0
				je		pcm86_end
				cmp		pcm86.fifo, 0
				jns		pcm86_end						// FIFOưƤʤ

				pushad
				mov		edi, ecx
				mov		bh, pcm86.dactrl
				test	bh, 30h
				je		pcm86_end2
				mov		cx, pcm86.bufsize
				movzx	esi, pcm86.readpos
				mov		bl, pcm86.count
				mov		ebp, 7fffh

				test	bh, 40h
				jne		pcm86_8bit

				test	bh, 20h
				je		pcm16bit_right
				test	bh, 10h
				je		pcm16bit_left

				// 16bit stereo out
				cmp		cx, 4
				jc		pcm86_end2
p86st16lp:		mov		ah, pcm86.buffer[esi]
				inc		esi
				and		esi, ebp
				mov		al, pcm86.buffer[esi]
				inc		esi
				and		esi, ebp
				cwde
				sar		eax, 1
				add		[edi], eax
				mov		ah, pcm86.buffer[esi]
				inc		esi
				and		esi, ebp
				mov		al, pcm86.buffer[esi]
				sub		esi, 3
				and		esi, ebp
				cwde
				sar		eax, 1
				add		[edi+4], eax
				add		edi, 8

				add		bl, pcm86.step
				cmp		bl, 32
				jc		p86st16incd
p86st16inc:		sub		cx, 4
				jc		p86st16lped2
				add		esi, 4
				and		esi, ebp
				sub		bl, 32
				cmp		bl, 32
				jnc		p86st16inc

p86st16incd:	dec		edx
				je		p86st16lped
				cmp		cx, 4
				jnc		p86st16lp

p86st16lped2:	and		cx, 3
p86st16lped:	mov		pcm86.readpos, si
				mov		pcm86.bufsize, cx
				mov		pcm86.count, bl
pcm86_end2:		popad
pcm86_end:		ret


				// 16bit mono out
pcm16bit_right:	add		edi, 4
pcm16bit_left:	cmp		cx, 2
				jc		p86mn16end
p86mn16lp:		mov		ah, pcm86.buffer[esi]
				inc		esi
				and		esi, ebp
				mov		al, pcm86.buffer[esi]
				dec		esi
				and		esi, ebp
				cwde
				sar		eax, 1
				add		[edi], eax
				add		edi, 8

				add		bl, pcm86.step
				cmp		bl, 32
				jc		p86mn16incd
				xor		eax, eax
				mov		al, bl
				shr		eax, 4
				and		eax, 0feh
				sub		cx, ax
				jc		p86mn16lped2
				add		esi, eax
				and		esi, ebp
				and		bl, 1fh
p86mn16incd:	dec		edx
				je		p86mn16lped
				cmp		cx, 2
				jnc		p86mn16lp
p86mn16lped2:	and		cx, 1
p86mn16lped:	mov		pcm86.readpos, si
				mov		pcm86.bufsize, cx
				mov		pcm86.count, bl
p86mn16end:		popad
				ret



pcm86_8bit:		test	bh, 20h
				je		pcm8bit_right
				test	bh, 10h
				je		pcm8bit_left

				// 8bit stereo out
				cmp		cx, 2
				jc		p86st8end
p86st8lp:		xor		eax, eax
				mov		ah, pcm86.buffer[esi]
				inc		esi
				and		esi, ebp
				cwde
				sar		eax, 1
				add		[edi], eax
				xor		eax, eax
				mov		ah, pcm86.buffer[esi]
				dec		esi
				and		esi, ebp
				cwde
				sar		eax, 1
				add		[edi+4], eax
				add		edi, 8

				add		bl, pcm86.step
				cmp		bl, 32
				jc		p86st8incd
p86st8inc:		sub		cx, 2
				jc		p86st8lped2
				add		esi, 2
				and		esi, ebp
				sub		bl, 32
				cmp		bl, 32
				jnc		p86st8inc

p86st8incd:		dec		edx
				je		p86st8lped
				cmp		cx, 2
				jnc		p86st8lp
p86st8lped2:	and		cx, 1
p86st8lped:		mov		pcm86.readpos, si
				mov		pcm86.bufsize, cx
				mov		pcm86.count, bl
p86st8end:		popad
				ret


				// 8bit mono out
pcm8bit_right:	add		edi, 4
pcm8bit_left:	cmp		cx, 0
				je		p86mn8end
p86mn8lp:		mov		ah, pcm86.buffer[esi]
				cwde
				sar		eax, 1
				add		[edi], eax
				add		edi, 8

				add		bl, pcm86.step
				cmp		bl, 32
				jc		p86mn8incd
				xor		eax, eax
				mov		al, bl
				shr		eax, 5
				sub		cx, ax
				jc		p86mn8lped2
				add		esi, eax
				and		esi, ebp
				and		bl, 1fh
p86mn8incd:		dec		edx
				je		p86mn8lped
				cmp		cx, 1
				jnc		p86mn8lp
p86mn8lped2:	mov		cx, 0
p86mn8lped:		mov		pcm86.readpos, si
				mov		pcm86.bufsize, cx
				mov		pcm86.count, bl
p86mn8end:		popad
				ret
		}
}

#else

INLINE void pcm86_monoclr(void) {

	pcm86.divremain = 0;
	pcm86.smp = 0;
	pcm86.lastsmp = 0;
}

INLINE void pcm86_stereoclr(void) {

	pcm86.divremain = 0;
	pcm86.smp_l = 0;
	pcm86.smp_r = 0;
	pcm86.lastsmp_l = 0;
	pcm86.lastsmp_r = 0;
}

static void pcmupdate_mono16(long *bufp, int buflength) {

	if (pcm86.div < PCM86_DIVENV) {					// åפפ
		do {
			long smp;
			if (pcm86.divremain < 0) {
				long dat;
				pcm86.divremain += PCM86_DIVENV;
				if (pcm86.bufsize < 2) {
					pcm86_monoclr();
					return;
				}
				pcm86.bufsize -= 2;
				dat = (char)pcm86.buffer[pcm86.readpos & 0x7fff] << 8;
				pcm86.readpos++;
				dat += pcm86.buffer[pcm86.readpos & 0x7fff];
				pcm86.readpos++;
				pcm86.lastsmp = pcm86.smp;
				pcm86.smp = dat;
			}
			smp = (pcm86.lastsmp * pcm86.divremain) -
							(pcm86.smp * (pcm86.divremain - PCM86_DIVENV));
			bufp[0] += (smp >> PCM86_DIVBIT);
			bufp += 2;
			pcm86.divremain -= pcm86.div;
		} while(--buflength);
	}
	else {
		do {
			long smp = pcm86.smp * (pcm86.divremain * -1);
			pcm86.divremain += PCM86_DIVENV;
			while(1) {
				long dat;
				if (pcm86.bufsize < 2) {
					pcm86_monoclr();
					return;
				}
				pcm86.bufsize -= 2;
				dat = (char)pcm86.buffer[pcm86.readpos & 0x7fff] << 8;
				pcm86.readpos++;
				dat += pcm86.buffer[pcm86.readpos & 0x7fff];
				pcm86.readpos++;
				pcm86.lastsmp = pcm86.smp;
				pcm86.smp = dat;
				if (pcm86.divremain > pcm86.div2) {
					pcm86.divremain -= pcm86.div2;
					smp += pcm86.smp * pcm86.div2;
				}
				else {
					break;
				}
			}
			smp += pcm86.smp * pcm86.divremain;
			bufp[0] += (smp >> PCM86_DIVBIT);
			bufp += 2;
			pcm86.divremain -= pcm86.div2;
		} while(--buflength);
	}
}

static void pcmupdate_stereo16(long *bufp, int buflength) {

	if (pcm86.div < PCM86_DIVENV) {					// åפפ
		do {
			long smp_l, smp_r;
			if (pcm86.divremain < 0) {
				long dat_l, dat_r;
				pcm86.divremain += PCM86_DIVENV;
				if (pcm86.bufsize < 4) {
					pcm86_stereoclr();
					return;
				}
				pcm86.bufsize -= 4;
				dat_l = (char)pcm86.buffer[pcm86.readpos & 0x7fff] << 8;
				pcm86.readpos++;
				dat_l += pcm86.buffer[pcm86.readpos & 0x7fff];
				pcm86.readpos++;
				pcm86.lastsmp_l = pcm86.smp_l;
				pcm86.smp_l = dat_l;

				dat_r = (char)pcm86.buffer[pcm86.readpos & 0x7fff] << 8;
				pcm86.readpos++;
				dat_r += pcm86.buffer[pcm86.readpos & 0x7fff];
				pcm86.readpos++;
				pcm86.lastsmp_r = pcm86.smp_r;
				pcm86.smp_r = dat_r;
			}
			smp_l = (pcm86.lastsmp_l * pcm86.divremain) -
							(pcm86.smp_l * (pcm86.divremain - PCM86_DIVENV));
			bufp[0] += (smp_l >> PCM86_DIVBIT);
			smp_r = (pcm86.lastsmp_r * pcm86.divremain) -
							(pcm86.smp_r * (pcm86.divremain - PCM86_DIVENV));
			bufp[1] += (smp_r >> PCM86_DIVBIT);
			bufp += 2;
			pcm86.divremain -= pcm86.div;
		} while(--buflength);
	}
	else {
		do {
			long smp_l, smp_r;
			smp_l = pcm86.smp_l * (pcm86.divremain * -1);
			smp_r = pcm86.smp_r * (pcm86.divremain * -1);
			pcm86.divremain += PCM86_DIVENV;
			while(1) {
				long dat_l, dat_r;
				if (pcm86.bufsize < 4) {
					pcm86_stereoclr();
					return;
				}
				pcm86.bufsize -= 4;
				dat_l = (char)pcm86.buffer[pcm86.readpos & 0x7fff] << 8;
				pcm86.readpos++;
				dat_l += pcm86.buffer[pcm86.readpos & 0x7fff];
				pcm86.readpos++;
				pcm86.lastsmp_l = pcm86.smp_l;
				pcm86.smp_l = dat_l * 256;

				dat_r = (char)pcm86.buffer[pcm86.readpos & 0x7fff] << 8;
				pcm86.readpos++;
				dat_r += pcm86.buffer[pcm86.readpos & 0x7fff];
				pcm86.readpos++;
				pcm86.lastsmp_r = pcm86.smp_r;
				pcm86.smp_r = dat_r * 256;

				if (pcm86.divremain > pcm86.div2) {
					pcm86.divremain -= pcm86.div2;
					smp_l += pcm86.smp_l * pcm86.div2;
					smp_r += pcm86.smp_r * pcm86.div2;
				}
				else {
					break;
				}
			}
			smp_l += pcm86.smp_l * pcm86.divremain;
			smp_r += pcm86.smp_r * pcm86.divremain;
			bufp[0] += (smp_l >> PCM86_DIVBIT);
			bufp[1] += (smp_r >> PCM86_DIVBIT);
			bufp += 2;
			pcm86.divremain -= pcm86.div2;
		} while(--buflength);
	}
}


static void pcmupdate_mono8(long *bufp, int buflength) {

	if (pcm86.div < PCM86_DIVENV) {					// åפפ
		do {
			long smp;
			if (pcm86.divremain < 0) {
				long dat;
				pcm86.divremain += PCM86_DIVENV;
				if (!pcm86.bufsize) {
					pcm86_monoclr();
					return;
				}
				pcm86.bufsize--;
				dat = (char)pcm86.buffer[pcm86.readpos & 0x7fff];
				pcm86.readpos++;
				pcm86.lastsmp = pcm86.smp;
				pcm86.smp = dat * 256;
			}
			smp = (pcm86.lastsmp * pcm86.divremain) -
							(pcm86.smp * (pcm86.divremain - PCM86_DIVENV));
			bufp[0] += (smp >> PCM86_DIVBIT);
			bufp += 2;
			pcm86.divremain -= pcm86.div;
		} while(--buflength);
	}
	else {
		do {
			long smp = pcm86.smp * (pcm86.divremain * -1);
			pcm86.divremain += PCM86_DIVENV;
			while(1) {
				long dat;
				if (!pcm86.bufsize) {
					pcm86_monoclr();
					return;
				}
				pcm86.bufsize--;
				dat = (char)pcm86.buffer[pcm86.readpos & 0x7fff];
				pcm86.readpos++;
				pcm86.lastsmp = pcm86.smp;
				pcm86.smp = dat * 256;
				if (pcm86.divremain > pcm86.div2) {
					pcm86.divremain -= pcm86.div2;
					smp += pcm86.smp * pcm86.div2;
				}
				else {
					break;
				}
			}
			smp += pcm86.smp * pcm86.divremain;
			bufp[0] += (smp >> PCM86_DIVBIT);
			bufp += 2;
			pcm86.divremain -= pcm86.div2;
		} while(--buflength);
	}
}

static void pcmupdate_stereo8(long *bufp, int buflength) {

	if (pcm86.div < PCM86_DIVENV) {					// åפפ
		do {
			long smp_l, smp_r;
			if (pcm86.divremain < 0) {
				long dat_l, dat_r;
				pcm86.divremain += PCM86_DIVENV;
				if (pcm86.bufsize < 2) {
					pcm86_stereoclr();
					return;
				}
				pcm86.bufsize -= 2;
				dat_l = (char)pcm86.buffer[pcm86.readpos & 0x7fff];
				pcm86.readpos++;
				pcm86.lastsmp_l = pcm86.smp_l;
				pcm86.smp_l = dat_l * 256;

				dat_r = (char)pcm86.buffer[pcm86.readpos & 0x7fff];
				pcm86.readpos++;
				pcm86.lastsmp_r = pcm86.smp_r;
				pcm86.smp_r = dat_r * 256;
			}
			smp_l = (pcm86.lastsmp_l * pcm86.divremain) -
							(pcm86.smp_l * (pcm86.divremain - PCM86_DIVENV));
			bufp[0] += (smp_l >> PCM86_DIVBIT);
			smp_r = (pcm86.lastsmp_r * pcm86.divremain) -
							(pcm86.smp_r * (pcm86.divremain - PCM86_DIVENV));
			bufp[1] += (smp_r >> PCM86_DIVBIT);
			bufp += 2;
			pcm86.divremain -= pcm86.div;
		} while(--buflength);
	}
	else {
		do {
			long smp_l, smp_r;
			smp_l = pcm86.smp_l * (pcm86.divremain * -1);
			smp_r = pcm86.smp_r * (pcm86.divremain * -1);
			pcm86.divremain += PCM86_DIVENV;
			while(1) {
				long dat_l, dat_r;
				if (pcm86.bufsize < 2) {
					pcm86_stereoclr();
					return;
				}
				pcm86.bufsize -= 2;
				dat_l = (char)pcm86.buffer[pcm86.readpos & 0x7fff];
				pcm86.readpos++;
				pcm86.lastsmp_l = pcm86.smp_l;
				pcm86.smp_l = dat_l * 256;

				dat_r = (char)pcm86.buffer[pcm86.readpos & 0x7fff];
				pcm86.readpos++;
				pcm86.lastsmp_r = pcm86.smp_r;
				pcm86.smp_r = dat_r * 256;

				if (pcm86.divremain > pcm86.div2) {
					pcm86.divremain -= pcm86.div2;
					smp_l += pcm86.smp_l * pcm86.div2;
					smp_r += pcm86.smp_r * pcm86.div2;
				}
				else {
					break;
				}
			}
			smp_l += pcm86.smp_l * pcm86.divremain;
			smp_r += pcm86.smp_r * pcm86.divremain;
			bufp[0] += (smp_l >> PCM86_DIVBIT);
			bufp[1] += (smp_r >> PCM86_DIVBIT);
			bufp += 2;
			pcm86.divremain -= pcm86.div2;
		} while(--buflength);
	}
}


void SOUND_CALL pcm86_update(void *buf, int buflength) {

	if ((buflength) && (pcm86.fifo & 0x80) && (pcm86.div)) {
		switch(pcm86.dactrl & 0x70) {
			case 0x00:						// 16bit-none
				break;

			case 0x10:						// 16bit-right
				pcmupdate_mono16(((long *)buf) + 1, buflength);
				break;

			case 0x20:						// 16bit-left
				pcmupdate_mono16(((long *)buf), buflength);
				break;

			case 0x30:						// 16bit-stereo
				pcmupdate_stereo16((long *)buf, buflength);
				break;

			case 0x40:						// 8bit-none
				break;

			case 0x50:						// 8bit-right
				pcmupdate_mono8(((long *)buf) + 1, buflength);
				break;

			case 0x60:						// 8bit-left
				pcmupdate_mono8((long *)buf, buflength);
				break;

			case 0x70:						// 8bit-stereo
				pcmupdate_stereo8((long *)buf, buflength);
				break;
		}
	}
}

#endif
