#include	"compiler.h"

#include	"np2.h"
#include	"rhythm.h"
#include	"np2sound.h"
#include	"dosio.h"
#include	"wave.h"

	RHYTHM_TABLE	rhythm;

static int waveread(RHYTHMFILE_TABLE *item, LPSTR filename, DWORD samplerate) {

	FILEH			src;
	WAVE_HEADER		whead;
	WAVE_INFOS		winfo;
	long			seekp;
	int				step;
	MEM_HDL			hwave;
	BYTE			*wave;
	DWORD			i;
	DWORD			size;
	DWORD			rate;
	DWORD			channels;
	DWORD			bit;

	if ((src = file_open_c(filename)) == (FILEH)-1) {
		return(FAILURE);
	}
	if ((file_read(src, &whead, sizeof(whead)) != sizeof(whead)) ||
		(whead.head != C4_DWORD('R','I','F','F'))) {
		file_close(src);
		return(FAILURE);
	}
	if ((file_read(src, &whead, 4) != 4) ||
		(whead.head != C4_DWORD('W','A','V','E'))) {
		file_close(src);
		return(FAILURE);
	}
	if ((file_read(src, &whead, sizeof(whead)) != sizeof(whead)) ||
		(whead.head != C4_DWORD('f','m','t',' ')) ||
		(file_read(src, &winfo, sizeof(winfo)) != sizeof(winfo))) {
		file_close(src);
		return(FAILURE);
	}
	seekp = GETDWORD(whead.size) + 20;
	file_seek(src, seekp, FSEEK_SET);
	if (file_read(src, &whead, sizeof(whead)) != sizeof(whead)) {
		file_close(src);
		return(FAILURE);
	}
	if (whead.head == C4_DWORD('f','a','c','t')) {
		seekp += GETDWORD(whead.size) + 8;
		file_seek(src, seekp, FSEEK_SET);
		if (file_read(src, &whead, sizeof(whead)) != sizeof(whead)) {
			file_close(src);
			return(FAILURE);
		}
	}
	if (whead.head != C4_DWORD('d','a','t','a')) {
		file_close(src);
		return(FAILURE);
	}

	rate = GETDWORD(winfo.rate);
	channels = GETWORD(winfo.channel);
	bit = GETWORD(winfo.bit);

	// եޥåȥå
	if ((rate != 44100) || (channels != 1) || (bit != 16)) {
		file_close(src);
		return(FAILURE);
	}

	size = GETDWORD(whead.size);
	step = rate / samplerate;
	item->rhythm_samples = size / (bit / 8) / channels / step;

	// waveɤ߹ߥ򥢥
	if ((hwave = np2os_malloc(size)) == NULL) {
		file_close(src);
		return(FAILURE);
	}
	wave = (BYTE *)MEM_PTR(hwave);

	// waveǡɤ߹
	if (file_lread(src, wave, size) != size) {
		np2os_free(hwave);
		file_close(src);
	}
	file_close(src);

	// ǡΥ
	if ((item->handle = np2os_malloc(item->rhythm_samples * 4)) == NULL) {
		np2os_free(hwave);
		return(FAILURE);
	}
	item->rhythm_data = (long *)MEM_PTR(item->handle);

	// ǡΥԡ
	for (i=0; i<item->rhythm_samples; i++) {
		item->rhythm_data[i] = (short)GETWORD(wave);
		wave += 2;
	}

	// 
	np2os_free(hwave);
	return(SUCCESS);
}


static char *rhythmfile[RHYTHM_MAX] = {
						"2608_bd.wav",
						"2608_sd.wav",
						"2608_top.wav",
						"2608_hh.wav",
						"2608_tom.wav",
						"2608_rim.wav"};


void rhythm_tableinit(void) {

	int		i;

	ZeroMemory(&rhythm, sizeof(rhythm));

	for (i=0; i<RHYTHM_MAX; i++) {
		waveread(&rhythm.r[i], rhythmfile[i], np2sound.rate);
	}
}


void rhythm_tableterm(void) {

	int		i;

	for (i=0; i<RHYTHM_MAX; i++) {
		if (rhythm.r[i].handle) {
			np2os_free(rhythm.r[i].handle);
			rhythm.r[i].handle = NULL;
		}
	}
}


void rhythm_reset(void) {

	int		i;

	for (i=0; i<6; i++) {
		rhythm.r[i].rhythm_volume = 0;
		rhythm.r[i].rhythm_lr = 0xc0;
	}
	rhythm.bitmap = 0;
}


void rhythm_volume(void) {									// ver0.28

	int		i;

	for (i=0; i<6; i++) {
		rhythm.r[i].rhythm_volume = 
					rhythm.r[i].rhythm_volreg * np2cfg.vol_rhythm;
	}
}


static void rhythm_play(int num, int event) {

	if ((num < RHYTHM_MAX) && (rhythm.r[num].handle)) {
		if (!event) {
			rhythm.bitmap |= (1 << num);
			rhythm.r[num].rhythm_ptr = rhythm.r[num].rhythm_data;
			rhythm.r[num].rhythm_remain = rhythm.r[num].rhythm_samples;
		}
		else {
			rhythm.bitmap &= ~(1 << num);
		}
	}
}

void rhythm_setreg(BYTE reg, BYTE value) {

	BYTE	bit;
	int		i;

	if (reg == 0x10) {
		sound_make();
		for (i=0, bit=1; i<6; i++, bit<<=1) {
			if (value & bit) {
				rhythm_play(i, value & 0x80);
			}
		}
	}
	else if ((reg >= 0x18) && (reg < 0x1e)) {
		rhythm.r[reg-0x18].rhythm_volreg = (value & 0x1f);		// ver0.28
		rhythm.r[reg-0x18].rhythm_volume = (value & 0x1f) * np2cfg.vol_rhythm;
		rhythm.r[reg-0x18].rhythm_lr = value & 0xc0;
	}
}


#ifdef SOUND_RHYTHM_ASM

LABEL void __fastcall rhythm_update(void *buf, int buflength) {

		__asm {
				cmp		edx, 0
				je		rhythm_end
				cmp		rhythm.bitmap, 0
				je		rhythm_end
				pushad
				xor		bx, bx
				mov		edi, offset (rhythm.r[0])
rhythm_lp:		bt		rhythm.bitmap, bx
				jnc		rhythm_next
				mov		ebp, [edi]RHYTHMFILE_TABLE.rhythm_remain
				cmp		edx, ebp
				jae		rhythm_remainad
				mov		ebp, edx
rhythm_remainad:sub		[edi]RHYTHMFILE_TABLE.rhythm_remain, ebp
				jg		rhythm_outst
				btr		rhythm.bitmap, bx
rhythm_outst:	mov		esi, [edi]RHYTHMFILE_TABLE.rhythm_ptr
				push	ecx
				push	edx
rhythm_outlp:	mov		eax, [esi]
				imul	[edi]RHYTHMFILE_TABLE.rhythm_volume
				sar		eax, 6 + 6								// ver0.27
				test	[edi]RHYTHMFILE_TABLE.rhythm_lr, 80h
				je		rhythm_outr
				add		[ecx], eax
rhythm_outr:	test	[edi]RHYTHMFILE_TABLE.rhythm_lr, 40h
				je		rhythm_outlped
				add		[ecx+4], eax
rhythm_outlped:	add		esi, 4
				add		ecx, 8
				dec		ebp
				jne		rhythm_outlp
				pop		edx
				pop		ecx
				mov		[edi]RHYTHMFILE_TABLE.rhythm_ptr, esi
rhythm_next:	add		edi, (type RHYTHMFILE_TABLE)
				inc		bx
				cmp		bx, RHYTHM_MAX
				jc		rhythm_lp
				popad
rhythm_end:		ret
		}
}

#else

void SOUND_CALL rhythm_update(void *buf, int buflength) {

	long				*p;
	WORD				bit;
	int					remain;
	long				*datp;
	long				samp;
	RHYTHMFILE_TABLE	*rchp;

	if ((!buflength) || (!rhythm.bitmap)) {
		return;
	}
	for (bit=1, rchp = rhythm.r; bit & 0x3f; bit<<=1, rchp++) {
		if (rhythm.bitmap & bit) {
			p = (long *)buf;
			remain = rchp->rhythm_remain;
			if (remain > buflength) {
				remain = buflength;
				rchp->rhythm_remain -= buflength;
			}
			else {
				rhythm.bitmap ^= bit;
				rchp->rhythm_remain = 0;
			}
			datp = rchp->rhythm_ptr;
			do {
				samp = *datp++;
				samp *= rchp->rhythm_volume;
				samp /= (1 << 12);
				if (rchp->rhythm_lr & 0x80) {
					p[0] += samp;
				}
				if (rchp->rhythm_lr & 0x40) {
					p[1] += samp;
				}
				p += 2;
			} while(--remain);
			rchp->rhythm_ptr = datp;
		}
	}
}

#endif
