#include	"compiler.h"

#include	"np2.h"
#include	"np2pc98.h"
#include	"i286.h"
#include	"dosio.h"
#include	"pc_core.h"
#include	"trace.h"
#include	"memory.h"

#include	"iobridge.h"
#include	"pic.h"
#include	"mouseif.h"
#include	"nevent.h"
#include	"gdc.h"
#include	"palettes.h"

	BYTE	np2trace_sw = 0;

#define	ENABLE_PCWATCH	0 //(np2trace_sw)
#define	ENABLE_INPORT	0 //(np2trace_sw)
#define	ENABLE_OUTPORT	0 //(np2trace_sw)
#define	ENABLE_OUTPORTW	0 //(np2trace_sw)
#define	ENABLE_STRING	1 //(np2trace_sw)
#define	ENABLE_MEM		0 //(np2trace_sw)
#define	ENABLE_VRAM		0 //(np2trace_sw)


static char *flagstr[16][2] = {
				{NULL, NULL},		// 0x8000
				{NULL, NULL},		// 0x4000
				{NULL, NULL},		// 0x2000
				{NULL, NULL},		// 0x1000
				{"NV", "OV"},		// 0x0800
				{"DN", "UP"},		// 0x0400
				{"DI", "EI"},		// 0x0200
				{NULL, NULL},		// 0x0100
				{"PL", "NG"},		// 0x0080
				{"NZ", "ZR"},		// 0x0040
				{NULL, NULL},		// 0x0020
				{"NA", "AC"},		// 0x0010
				{NULL, NULL},		// 0x0008
				{"PO", "PE"},		// 0x0004
				{NULL, NULL},		// 0x0002
				{"NC", "CY"}};		// 0x0001

static char *flags(void) {

static char work[512];
		int		i;
		WORD	bit;
		WORD	flag;
		char	w2[256];

	wsprintf(work, 
			"AX=%04x  BX=%04x  CX=%04x  DX=%04x  "		\
			"SP=%04x  BP=%04x  SI=%04x  DI=%04x\n"		\
			"DS=%04x  ES=%04x  SS=%04x  CS=%04x  "		\
			"IP=%04x   ",
			I286_AX, I286_BX, I286_CX, I286_DX,
			I286_SP, I286_BP, I286_SI, I286_DI,
			I286_DS, I286_ES, I286_SS, I286_CS,
			I286_IP);

	flag = I286_FLAG;
	for (i=0, bit=0x8000; bit; i++, bit>>=1) {
		if (flagstr[i][0]) {
			if (flag & bit) {
				strcat(work, flagstr[i][1]);
			}
			else {
				strcat(work, flagstr[i][0]);
			}
			if (bit != 1) {
				strcat(work, " ");
			}
		}
	}
	strcat(work, "\r\n");

	wsprintf(w2, "DS=%08x ES=%08x SS=%08x CS=%08x\r\n",
									DS_BASE, ES_BASE, SS_BASE, CS_BASE);
	strcat(work, w2);

	return(work);
}





#ifdef TRACE

static	FILEH		dfh = (FILEH)-1;

#define TRACEMAX	(1 << 14)
static	WORD	trace_SEG[TRACEMAX];
static	WORD	trace_OFF[TRACEMAX];
static	int		trace_pos = 0;

void tracelog_start(void) {

	dfh = file_create_c("tracelog.txt", FTYPE_TEXT);
}

void tracelog_end(void) {

	if (dfh != (FILEH)-1) {
		file_close(dfh);
	}
}

void pcwatch(void) {

	char	work[128];

	if (I286_CS != 0x1fc0) {
		return;
	}
	if (ENABLE_PCWATCH) {
		if (dfh != (FILEH)-1) {
#if 1
			char *p = flags();
			file_write(dfh, p, strlen(p));
#endif
			wsprintf(work, "%04x:%04x\r\n", I286_CS, I286_IP);
			file_write(dfh, work, strlen(work));
		}
	}
}

void trace_inport(WORD a) {

	char	work[64];

	if (ENABLE_INPORT) {
		if (dfh != (FILEH)-1) {
			wsprintf(work, "in  %04x [%04x:%04x]\r\n", a, I286_CS, I286_IP);
			file_write(dfh, work, strlen(work));
		}
	}
}

void trace_inport_a(WORD a, BYTE b) {

	char	work[64];

	if (ENABLE_INPORT) {
		if (dfh != (FILEH)-1) {
			wsprintf(work, "in  %04x, %02x [%04x:%04x]\r\n", a, b,
														I286_CS, I286_IP);
			file_write(dfh, work, strlen(work));
		}
	}
}

void trace_outport(WORD a, BYTE b) {

	char	work[64];

	if (ENABLE_OUTPORT) {
		if (dfh != (FILEH)-1) {
#if 0
			wsprintf(work, "out %04x,%02x - %04x:%04x\r\n",
									a, b, I286_CS, I286_IP);
#else
			wsprintf(work, "out %04x,%04x - %10d\r\n",
									a, b, 
				(nevent.clock + nevent.baseclock - nevent.remainclock));
#endif
			file_write(dfh, work, strlen(work));
		}
	}
}

void trace_outport_w(WORD a, WORD b) {

	char	work[64];

	if (ENABLE_OUTPORTW) {
		if (dfh != (FILEH)-1) {
			wsprintf(work, "out %04x,%04x - %04x:%04x\r\n",
									a, b, I286_CS, I286_IP);
			file_write(dfh, work, strlen(work));
		}
	}
}

void outstrbyte(char *str, WORD val) {

	char	work[512];

	if (ENABLE_STRING) {
		if (dfh != (FILEH)-1) {
			wsprintf(work, "%s %x\r\n", str, val);
			file_write(dfh, work, strlen(work));
		}
	}
}

#ifdef __cplusplus
extern "C" {
#endif

void MEM_CALL tracememwrite(DWORD adrs, BYTE data) {		// extrn C

	char	work[64];

#if 0
	if ((adrs < 0x1A800) || (adrs >= (0x1A800 + 0x1578*2))) {
		return;
	}
#endif
	if (ENABLE_MEM) {
		if (dfh != (FILEH)-1) {
			wsprintf(work, "memw %05x,%02x - %04x:%04x\r\n",
									adrs, data, I286_CS, I286_IP);
			file_write(dfh, work, strlen(work));
		}
	}
}

void MEM_CALL tracememwrite_w(DWORD adrs, WORD data) {	// extrn C

	char	work[64];

#if 0
	if ((adrs < 0x1A800) || (adrs >= (0x1A800 + 0x1578*2))) {
		return;
	}
#endif
	if (ENABLE_MEM) {
		if (dfh != (FILEH)-1) {
			wsprintf(work, "memw %05x,%04x - %04x:%04x\r\n",
									adrs, data, I286_CS, I286_IP);
			file_write(dfh, work, strlen(work));
		}
	}
}



void MEM_CALL tracevramwrite(DWORD adrs, BYTE data) {		// extrn C

	char	work[512];

	if (ENABLE_VRAM) {
		if (dfh != (FILEH)-1) {
			wsprintf(work, "w %05x,%02x - %04x:%04x\r\n",
									adrs, data, I286_CS, I286_IP);
			file_write(dfh, work, strlen(work));
		}
	}
}

void MEM_CALL tracevramwrite_w(DWORD adrs, WORD data) {	// extrn C

	char	work[512];

	if (ENABLE_VRAM) {
		if (dfh != (FILEH)-1) {
			wsprintf(work, "w %05x,%04x - %04x:%04x\r\n",
									adrs, data, I286_CS, I286_IP);
			file_write(dfh, work, strlen(work));
		}
	}
}

void MEM_CALL tracevramread(DWORD adrs) {					// extrn C

	char	work[512];

	if (ENABLE_VRAM) {
		if (dfh != (FILEH)-1) {
			wsprintf(work, "r %05x - %04x:%04x\r\n",
									adrs, I286_CS, I286_IP);
			file_write(dfh, work, strlen(work));
		}
	}
}
#ifdef __cplusplus
}
#endif

#endif





// ----------------------------------------------------------------- interface

void np2trace_status(void) {

static	int		filenum = 0;
		FILEH	fh;
		char	work[512];
		char	*p;

	wsprintf(work, "i286reg.%03d", filenum);
	if ((fh = file_create_c(work, FTYPE_TEXT)) != (FILEH)-1) {
		p = flags();
		file_write(fh, p, strlen(p));

		wsprintf(work, "\nPIC0=%02x:%02x:%02x\nPIC1=%02x:%02x:%02x\n"
				"8255PORTC = %02x",
				pic[0].imr, pic[0].irr, pic[0].isr,
				pic[1].imr, pic[1].irr, pic[1].isr,
				mouseif.portc);

		file_write(fh, work, strlen(work));

		file_close(fh);
	}

	wsprintf(work, "i286_cs.%03d", filenum);
	if ((fh = file_create_c(work, FTYPE_TEXT)) != (FILEH)-1) {
		file_lwrite(fh, &mem[CS_BASE], 0x10000);
		file_close(fh);
	}
	wsprintf(work, "i286_ds.%03d", filenum);
	if ((fh = file_create_c(work, FTYPE_TEXT)) != (FILEH)-1) {
		file_lwrite(fh, &mem[DS_BASE], 0x10000);
		file_close(fh);
	}
	wsprintf(work, "i286_es.%03d", filenum);
	if ((fh = file_create_c(work, FTYPE_TEXT)) != (FILEH)-1) {
		file_lwrite(fh, &mem[ES_BASE], 0x10000);
		file_close(fh);
	}
	wsprintf(work, "i286_ss.%03d", filenum);
	if ((fh = file_create_c(work, FTYPE_TEXT)) != (FILEH)-1) {
		file_lwrite(fh, &mem[SS_BASE], 0x10000);
		file_close(fh);
	}
	wsprintf(work, "i286_txt.%03d", filenum);
	if ((fh = file_create_c(work, FTYPE_TEXT)) != (FILEH)-1) {
		file_lwrite(fh, &mem[0xa0000], 0x4000);
		file_close(fh);
	}

#if 0
	wsprintf(work, "EGC%03d.LOG", filenum);
	if ((fh = file_create_c(work, FTYPE_TEXT)) != (FILEH)-1) {
		int		cnt;
		char	bufs[256];
		extern	int		egcropcnt[];
		for (cnt=0; cnt<256; cnt++) {
			if (egcropcnt[cnt]) {
				wsprintf(bufs, "rop:%02x - %8u\r\n", cnt, egcropcnt[cnt]);
				file_write(fh, bufs, strlen(bufs));
			}
		}
		file_close(fh);
	}
#endif

	filenum++;
}


void np2trace_memorydump(void) {

	FILEH	fh;
	int		i;

	if ((fh = file_create_c("memory.bin", FTYPE_NONE)) != (FILEH)-1) {
		for (i=0; i<34; i++) {
			file_write(fh, mem + i*0x8000, 0x8000);
		}
		file_close(fh);
	}
}

