/*	$Id: np2osdep.c,v 1.13 2003/06/05 18:10:17 nonaka Exp $	*/

/*
 * Copyright (c) 2001-2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "compiler.h"

#include "np2.h"
#include "audio.h"
#include "dosio.h"
#include "midi.h"
#include "timing.h"

#include "xnp2.h"
#include "xkeyboard.h"
#include "xmouse.h"

static void
sound_init(void)
{

	audio_setup();
	audio_init(np2cfg.samplingrate, np2cfg.delayms);
	audio_play();
}

static void
sound_term(void)
{

	audio_stop();
	audio_term();
}

void
np2os_init(void)
{

	xkeyboard_init(window);
	xmouse_init(window);
	audio_setup();
	MIDI_Init();
}

void
np2os_reset(void)
{

	MIDI_Cleanup();
	sound_term();
	sound_init();
	MIDI_Init();

	timing_init();
}

void
np2os_term(void)
{

	sound_term();
	xmouse_term(window);
}


void
np2os_getcalender(CALENDER_T *c)
{
	time_t ltime;
	struct tm *nt;

	time(&ltime);
	nt = localtime(&ltime);

	c->year = nt->tm_year;
	c->month = nt->tm_mon + 1;
	c->day = nt->tm_mday;
	c->hour = nt->tm_hour;
	c->minute = nt->tm_min;
	c->second = nt->tm_sec;
	c->weekday = nt->tm_wday;
}
